/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */
package com.funambol.framework.engine.pipeline;

import java.io.Serializable;
import java.util.Map;
import java.util.HashMap;

/**
 * This class allows the pipeline components to share session-scoped and 
 * request-scoped properties.
 * <p>
 * Standard session properties are:
 * <ul>
 *   funambol.session.id - the server session id
 * </ul>
 * Standard request properties are:
 * <ul>
 *   funambol.request.parameters - parameters specified on the query string of
 *                               the current request
 *   funambol.request.headers - headers given with the current message
 * </ul>
 * <p>
 * <b>NOTE:</b> the prefix <i>funambol.</i> for property keys is reserved for
 * Funambol internal use.
 *
 * @version $Id: MessageProcessingContext.java,v 1.1 2006/01/18 17:00:07 luigiafassina Exp $
 */
public class MessageProcessingContext implements Serializable {

    // --------------------------------------------------------------- Constants
    public static final String PROPERTY_SESSION_ID = "funambol.session.id";
    
    public static final String PROPERTY_REQUEST_HEADERS 
        = "funambol.request.headers";
    
    public static final String PROPERTY_REQUEST_PARAMETERS
        = "funambol.request.parameters";
    
    // ------------------------------------------------------------ Private data
    
    /**
     * Session scoped properties
     */
    private HashMap sessionProperties;
    
    /**
     * Request scoped properties
     */
     private HashMap requestProperties;
    
    // ------------------------------------------------------------ Constructors
    
    /**
     * Creates a new MessageProcessingContext object
     *
     */
    public MessageProcessingContext() {
        sessionProperties = new HashMap();
        requestProperties = new HashMap();
    }

    // ---------------------------------------------------------- Public methods

    /**
     * Getter for sessionProperties
     * @return Value of properties
     */
    public Map getSessionProperties() {
        return this.sessionProperties;
    }

    /**
     * Setter for sessionProperties.
     *
     * @param properties New value of properties.
     */
    public void setSessionProperties(Map properties) {
        this.sessionProperties.clear();
        this.sessionProperties.putAll(properties);
    }
    
    /**
     * Getter for requestProperties
     * @return Value of properties
     */
    public Map getRequestProperties() {
        return this.requestProperties;
    }

    /**
     * Setter for sessionProperties.
     *
     * @param properties New value of properties.
     */
    public void setRequestProperties(Map properties) {
        this.requestProperties.clear();
        this.requestProperties.putAll(properties);
    }
    
    /**
     * Returns the value of the session property associated to the given key
     *
     * @param key the context property key
     *
     * @return the property value or null if key was not found
     */
    public Object getSessionProperty(String key) {
        return sessionProperties.get(key);
    }

    /**
     * Sets a given session property
     *
     * @param key the context property key
     * @param value the context property value
     *
     */
    public void setSessionProperty(String key, Object value) {
        sessionProperties.put(key, value);
    }

    
    /**
     * Returns the value of the request property associated to the given key
     *
     * @param key the context property key
     *
     * @return the property value or null if key was not found
     */
    public Object getRequestProperty(String key) {
        return requestProperties.get(key);
    }

    /**
     * Sets a given request property
     *
     * @param key the context property key
     * @param value the context property value
     *
     */
    public void setRequestProperty(String key, Object value) {
        requestProperties.put(key, value);
    }
    
    /**
     * Returns a String representation of request and session properties 
     * (for debug purposes).
     *
     * @return a String representation of request and session properties 
     * (for debug purposes).
     */
    public String toString() {
        StringBuffer sb = new StringBuffer();
        
        sb.append(super.toString()).append(" {")
          .append("\n\tSession properties: ").append(sessionProperties)
          .append("\n\tRequest properties: ").append(requestProperties)
          .append("\n}");
        
        return sb.toString();
          
    }

}