/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package com.funambol.framework.engine.source;

import java.util.logging.Logger;
import java.util.logging.Level;
import java.security.Principal;
import java.sql.Timestamp;

import com.funambol.framework.logging.Sync4jLogger;
import com.funambol.framework.engine.*;
import com.funambol.framework.engine.source.SyncSource;
import com.funambol.framework.filter.FilterClause;

/**
 * Implements the basic functionalities of a <i>SyncSource</i> like naming.
 *
 *
 * @version $Id: AbstractSyncSource.java,v 1.1 2006/01/18 17:00:07 luigiafassina Exp $
 */
public abstract class AbstractSyncSource implements SyncSource, java.io.Serializable {
    // --------------------------------------------------------------- Constants

    public static final String LOG_NAME = "engine.source";

    // ---------------------------------------------------------- Protected data

    protected String name       = null;
    protected String type       = null;
    protected String sourceURI  = null;
    protected String sourceQuery= null;
    protected SyncSourceInfo info      = null;

    /**
     * How many items were added ?
     */
    protected long howManyAdded;

    /**
     * How many items were deleted ?
     */
    protected long howManyDeleted;

    /**
     * How many items were updated ?
     */
    protected long howManyUpdated;

    // ------------------------------------------------------------ Constructors

    /** Creates a new instance of AbstractSyncSource */
    protected AbstractSyncSource() {
    }

    public AbstractSyncSource(String name, String type, String sourceURI) {
        if (name == null) {
            throw new NullPointerException("name is null!");
        }

        this.name      = name;
        this.type      = (type == null) ? "unknown" : type;
        setSourceURI(sourceURI);
    }

    public AbstractSyncSource(String name) {
        this(name, null, null);
    }

    // ---------------------------------------------------------- Public methods

    public String getName() {
        return name;
    }

    public void setName(String name) {
        this.name = name;
    }

    public String getType() {
        return this.type;
    }

    public void setType(String type) {
        this.type = type;
    }

    /** Getter for property query.
     * @return Value of property query.
     */
    public String getSourceQuery() {
        return sourceQuery;
    }

    /** Getter for property uri.
     * @return Value of property uri.
     */
    public String getSourceURI() {
        return sourceURI;
    }

    /** Setter for property uri.
     * @param sourceURI New value of property uri.
     */
    public void setSourceURI(String sourceURI) {
        int qMark = sourceURI.indexOf('?');
        if (qMark == -1) {
            this.sourceURI = sourceURI;
            this.sourceQuery = "";
        } else {
            this.sourceURI = sourceURI.substring(0, qMark);
            this.sourceQuery = sourceURI.substring(qMark);
        }
    }

    /** Returns the type info of the content handled by this source
     *
     * @return the type info of the content handled by this source
     *
     */
    public SyncSourceInfo getInfo() {
        return this.info;
    }

    /**
     * Setter for the property <i>info</i>
     */
    public void setInfo(SyncSourceInfo info) {
        this.info = info;
    }

    /**
     * Returns a string representation of this object.
     *
     * @return a string representation of this object.
     */
    public String toString() {
        StringBuffer sb = new StringBuffer(super.toString());

        sb.append(" - {name: ").append(getName()     );
        sb.append(" type: "   ).append(getType()     );
        sb.append(" uri: "    ).append(getSourceURI());
        sb.append("}"         );

        return sb.toString();
    }

    /**
     * This resets the howMany counters. It should then be called by extending
     * classes before anything else.
     *
     * @param principal user/device
     * @param syncMode sync mode
     * @param filter the filter to use
     *
     * @throws SyncSourceException in case of any error
     *
     * @see SyncSource
     */
    public void beginSync(Principal    principal,
                          int          syncMode,
                          FilterClause filter) throws SyncSourceException {

        howManyAdded = howManyDeleted = howManyUpdated = 0;
    }

    /**
     * It logs the howMany counters. It should then be called by extending
     * classes before anything else.
     *
     * @param principal user/device
     *
     * @throws SyncSourceException in case of any error
     *
     * @see SyncSource
     */
    public void endSync() throws SyncSourceException {
        Logger log = Sync4jLogger.getLogger(LOG_NAME);

        if (log.isLoggable(Level.INFO)) {
            log.info(getSourceURI()
                    + ": "
                    + howManyAdded
                    + " new items added, "
                    + howManyUpdated
                    + " existing items updated, "
                    + howManyDeleted
                    + " items deleted."
                    );
        }
    }

    // -------------------------------------------------------- Abstract methods

    public abstract SyncItemKey[] getAllSyncItemKeys() throws SyncSourceException;

    public abstract SyncItemKey[] getDeletedSyncItemKeys(Timestamp sinceTs,
                                                         Timestamp untilTs)
        throws SyncSourceException;

    public abstract SyncItemKey[] getNewSyncItemKeys(Timestamp sinceTs,
                                                     Timestamp untilTs)
        throws SyncSourceException;


    public abstract SyncItem getSyncItemFromId(SyncItemKey syncItemKey) throws SyncSourceException;

    public abstract char getSyncItemStateFromId(SyncItemKey syncItemKey) throws SyncSourceException;


    public abstract void removeSyncItem(SyncItemKey syncItemKey,
                                        Timestamp   time,
                                        boolean     softDelete) throws SyncSourceException;

    public abstract void removeSyncItem(SyncItemKey syncItemKey,
                                        Timestamp time) throws SyncSourceException;

    public abstract SyncItem updateSyncItem(SyncItem syncInstance) throws SyncSourceException;

    public abstract SyncItem addSyncItem(SyncItem syncInstance) throws SyncSourceException;

    public abstract SyncItemKey[] getSyncItemKeysFromTwin(SyncItem syncItem) throws SyncSourceException;


    public abstract boolean mergeSyncItems(SyncItemKey serverItem,
                                           SyncItem    clientItem) throws SyncSourceException;

    public abstract boolean isSyncItemInFilterClause(SyncItem item) throws SyncSourceException;

    public abstract boolean isSyncItemInFilterClause(SyncItemKey key) throws SyncSourceException;

    public abstract void setOperationStatus(String operationName, int status, SyncItemKey[] keys);

}
