/**
 * Copyright (C) 2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.funambol.framework.engine.transformer;

import java.util.Map;
import com.funambol.framework.engine.SyncItem;
import java.util.StringTokenizer;
import com.funambol.framework.logging.Sync4jLogger;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.util.HashMap;
import com.funambol.framework.tools.beans.BeanFactory;

/**
 * This class represents a data encoder or decoder
 *
 * A DataTransformer is a component that given some data in some format
 * as input, it generates a transformation of the source data in output.
 * Examples of transformations are:
 *
 * - base64 encoding/decoding
 * - encryption/decription
 */
public class DataTransformerManager {

    // ------------------------------------------------------------ Private data
    private static final Logger log = Sync4jLogger.getLogger("engine");

    private Map sourceUriTrasformationsRequired;

    private Map dataTransformersIn;
    private Map dataTransformersOut;

    // ------------------------------------------------------------ Constructors


    /** Creates a new instance of DataTransformerManager */
    public DataTransformerManager() {
        sourceUriTrasformationsRequired = new HashMap();
        dataTransformersIn  = new HashMap();
        dataTransformersOut = new HashMap();
    }

    // ---------------------------------------------------------- Public methods

    public Map getSourceUriTrasformationsRequired() {
        return sourceUriTrasformationsRequired;
    }

    public void setSourceUriTrasformationsRequired(Map sourceUriTrasformationsRequired) {
        this.sourceUriTrasformationsRequired = sourceUriTrasformationsRequired;
    }

    public Map getDataTransformersIn() {
        return dataTransformersIn;
    }


    public void setDataTransformersIn(Map dataTransformersIn) {
        this.dataTransformersIn = dataTransformersIn;
    }


    public Map getDataTransformersOut() {
        return dataTransformersOut;
    }


    public void setDataTransformersOut(Map dataTransformersOut) {
        this.dataTransformersOut = dataTransformersOut;
    }


    public SyncItem transformIngoingItem(TransformationInfo info, SyncItem item)
    throws TransformerException {

        String format = item.getFormat();
        if (format == null) {
            //
            // No transformations are applicable
            //
            return item;
        }

        String uri    = item.getSyncSource().getSourceURI();

        String transformerNames = (String)sourceUriTrasformationsRequired.get(uri);
        if (transformerNames == null) {
            //
            // no transformations are required
            //
            return item;
        }

        if (format.endsWith(transformerNames)) {
            //
            // We have to apply the transformes specifies in the transformerNames
            // in inverse order
            //
            if (log.isLoggable(Level.FINEST)) {
                log.finest("Applying transformations '" + transformerNames + "'");
            }

            // textplain;des;b64
            byte[]          content  = item.getContent();
            StringTokenizer st       = new StringTokenizer(format, ";");
            int             numToken = st.countTokens();
            String[]        tokens   = new String[numToken];
            DataTransformer dt       = null;
            for (int i=tokens.length-1; i>=0; i--) {
                tokens[i] = st.nextToken();

                if (log.isLoggable(Level.FINEST)) {
                    log.finest("Applying transformation: '" + tokens[i] + "'");
                }

                dt = getDataTransformerIn(tokens[i]);

                if (dt != null) {
                    content = dt.transform(content, info);
                    if (log.isLoggable(Level.FINEST)) {
                        log.finest("Transformation '" + tokens[i] + "' done");
                    }
                } else {
                    if (log.isLoggable(Level.FINEST)) {
                        log.finest("DataTransformer for '" + tokens[i] +
                                   "' for item goes in not found");
                    }
                }
            }
            item.setContent(content);

            format = format.substring(0, format.indexOf(transformerNames));
            item.setFormat(format);
        }


        return item;
    }

    public SyncItem transformOutgoingItem(TransformationInfo info, SyncItem item)
    throws TransformerException {

        String format = item.getFormat();
        String uri    = item.getSyncSource().getSourceURI();

        String transformerNames = (String)sourceUriTrasformationsRequired.get(uri);
        if (transformerNames == null) {
            //
            // no transformations are required
            //
            return item;
        }

        //
        // We have to apply the transformes specifies in the transformerNames
        // in inverse order
        //
        if (log.isLoggable(Level.FINEST)) {
            log.finest("Applying transformations '" + transformerNames + "'");
        }

        // textplain;des;b64
        StringBuffer    sb       = new StringBuffer(format);
        byte[]          content  = item.getContent();
        StringTokenizer st       = new StringTokenizer(format, ";");
        int             numToken = st.countTokens();
        String[]        tokens   = new String[numToken];
        DataTransformer dt       = null;
        for (int i=0; i<tokens.length; i++) {
            tokens[i] = st.nextToken();
            if (log.isLoggable(Level.FINEST)) {
                log.finest("Applying transformation: '" + tokens[i] + "'");
            }

            dt = getDataTransformerOut(tokens[i]);

            if (dt != null) {
                content = dt.transform(content, info);
                if (sb.length() != 0) {
                    sb.append(";");
                }
                sb.append(tokens[i]);
                if (log.isLoggable(Level.FINEST)) {
                    log.finest("Transformation '" + tokens[i] + "' done");
                }
            } else {
                if (log.isLoggable(Level.FINEST)) {
                    log.finest("DataTransformer for '" + tokens[i] +
                               "' for item goes in not found");
                }
                throw new TransformerException("DataTransformer for '" + tokens[i] +
                                               "' for item goes in not found");
            }
        }
        item.setContent(content);
        item.setFormat(sb.toString());

        return item;
    }

    private DataTransformer getDataTransformerIn(String name)
    throws TransformerException {
        String          dtClass = (String)dataTransformersIn.get(name);
        DataTransformer dt      = null;

        if (log.isLoggable(Level.FINEST)) {
            log.finest("DataTransformer class for '" + name + "' in input: " + dt);
        }
        if (dtClass != null) {
            try {
                dt = (DataTransformer) Class.forName(dtClass).newInstance();
            } catch (Exception ex) {
                log.throwing(this.getClass().getName(), "getDataTransformerIn", ex);
                throw new TransformerException("Unable to create an instance of '" + dtClass + "'");
            }
        }
        if (log.isLoggable(Level.FINEST)) {
            log.finest("DataTransformer instance for '" + name + "' in input: " + dt);
        }
        return dt;
    }

    private DataTransformer getDataTransformerOut(String name)
    throws TransformerException {
        String          dtClass = (String)dataTransformersOut.get(name);
        DataTransformer dt      = null;

        if (log.isLoggable(Level.FINEST)) {
            log.finest("DataTransformer class for '" + name + "' in output: " + dt);
        }
        if (dtClass != null) {
            try {
                dt = (DataTransformer) Class.forName(dtClass).newInstance();
            } catch (Exception ex) {
                log.throwing(this.getClass().getName(), "getDataTransformerOut", ex);
                throw new TransformerException("Unable to create an instance of '" + dtClass + "'");
            }
        }
        if (log.isLoggable(Level.FINEST)) {
            log.finest("DataTransformer instance for '" + name + "' in output: " + dt);
        }
        return dt;
    }

    public static void main(String[] args) throws Exception {
        DataTransformerManager dm = new DataTransformerManager();

        Map sourceUriTrasformersIn = new HashMap();

        Map dataTransformersIn = new HashMap();
        Map dataTransformersOut = new HashMap();

        sourceUriTrasformersIn.put("scard", "b64");
        sourceUriTrasformersIn.put("scal", "b64");

        dataTransformersIn.put("b64", "com.funambol.server.engine.transformer.B64Decoder");
        dataTransformersOut.put("b64", "com.funambol.server.engine.transformer.B64Encoder");

        dm.setSourceUriTrasformationsRequired(sourceUriTrasformersIn);

        dm.setDataTransformersIn(dataTransformersIn);
        dm.setDataTransformersOut(dataTransformersOut);


        String xml = BeanFactory.marshal(null, dm);
        System.out.println("xml: " + xml);

    }

}
