/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.funambol.server.engine;

import java.security.Principal;
import java.sql.Timestamp;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.apache.commons.collections.ListUtils;

import com.funambol.framework.core.Add;
import com.funambol.framework.core.ModificationCommand;
import com.funambol.framework.core.StatusCode;
import com.funambol.framework.engine.*;
import com.funambol.framework.engine.Util;
import com.funambol.framework.engine.source.MemorySyncSource;
import com.funambol.framework.engine.source.ObjectSizeMismatchException;
import com.funambol.framework.engine.source.SyncSource;
import com.funambol.framework.engine.source.SyncSourceException;
import com.funambol.framework.server.ClientMapping;
import com.funambol.framework.core.Delete;
import com.funambol.framework.server.ClientMappingEntry;

/**
 * This class represents a synchronization process.
 *
 * The base concrete implementation of a synchronization strategy. It implements
 * the <i>ConcreteStrategy</i> partecipant of the Strategy Pattern.
 * <p>
 * The synchronization process is implemented in this class as follows:
 * <p>
 * Given a set of sources A, B, C, D, etc, the synchronization process takes place
 * between two sources at a time: A is first synchronzed with B, then AB with
 * C, then ABC with D and so on. <br>
 * The synchronization process is divided in three phases: preparation, synchronization,
 * finalization. These phases correspond to the methdos prepareSync, sync and
 * endSync; they are call-back methods called sequentially by a driver object,
 * usually a SyncMethod object. In this way the SyncMethod object has a chance
 * to communicate with the external world before carry on the process.
 * For example a SyncMethod object can show to the user the items that are going
 * to be synchronized, so that the user can let the process carry on or stop it. <br>
 * prepareSync returns an array of SyncOperation, in which each element represents
 * a particular synchronization action, ie. create an item in the source A,
 * delete the item X from the source B, etc. Sometime it is not possible decide
 * what to do, thus a SyncConflict operation is used. A conflict must be solved
 * by something external the synchronization process, for instance by a user
 * action. Below is a table of all possible situations.
 * <pre>
 *
 * | -------- | --- | --- | --- | --- | --- |
 * | Source A |     |     |     |     |     |
 * |    /     |  N  |  D  |  U  |  S  |  X  |     N : item new
 * | Source B |     |     |     |     |     |     D : item deleted
 * | -------- | --- | --- | --- | --- | --- |     U : item updated
 * |       N  |  O  |  O  |  O  |  O  |  B  |     S : item synchronized/unchanged
 * | -------- | --- | --- | --- | --- | --- |     X : item not existing
 * |       D  |  O  |  X  |  O  |  X  |  X  |     O : conflict
 * | -------- | --- | --- | --- | --- | --- |     A : item A replaces item B
 * |       U  |  O  |  O  |  O  |  B  |  B  |     B : item B replaces item A
 * | -------- | --- | --- | --- | --- | --- |
 * |       S  |  O  |  X  |  A  |  =  |  B  |
 * | -------- | --- | --- | --- | --- | --- |
 * |       X  |  A  |  X  |  A  |  A  |  X  |
 * | -------- | --- | --- | --- | --- | --- |
 *
 * </pre>
 *
 *
 * @version  $Id: Sync4jStrategy.java,v 1.2 2006/01/30 17:11:18 nichele Exp $
 */
public class Sync4jStrategy implements SyncStrategy, java.io.Serializable {

    // --------------------------------------------------------------- Constants

    public static String LOG_NAME = "funambol.framework.engine";

    private transient static Logger log = Logger.getLogger(LOG_NAME);

    // -------------------------------------------------------------- Properties

    /**
     * The synchronization source
     */
    private transient SyncSource[] sources = null;
    public SyncSource[] getSources() {
        return sources;
    }

    public void setSources(SyncSource[] sources) {
        this.sources = sources;
    }

    /**
     * The process' name
     */
    public String getName() {
        return name;
    }

    private String name;
    public void setName(String name) {
        this.name = name;
    }

    // ------------------------------------------------------------ Private data

    //
    // NOTE: server modification MUST be detected and stored at the beginning of
    // the sync process (i.e. in the first modification message), otherwise in
    // the case of multi message, likely the server will detect items modified
    // by the current sync as server modifications.
    // Reading server modifications at the synchronization start does not
    // loose items since. Let's suppose we have the following events axis:
    //
    //  s1            e1    s2              e2             s3          e3
    //  |-------------|....|---------------|...............|-----------|
    //
    // where s<x> is the beginning of a sync and e<x> represents its end.
    // In s2 we collect all modified items since s1; if any other process
    // modifies an item after s2, it won't be detected during the second
    // synchronization. However, it will be detected at s3, where we collect all
    // modified items since s2.
    //
    private HashMap newBAll;
    private HashMap updatedBAll;
    private HashMap deletedBAll;
    private HashMap BmAll; //used into Slow Sync

    // ------------------------------------------------------------ Constructors

    public Sync4jStrategy() {
        newBAll = new HashMap();
        updatedBAll = new HashMap();
        deletedBAll = new HashMap();
        BmAll = new HashMap();
    }

    public Sync4jStrategy(SyncSource[] syncSources) {
        this();
        sources = syncSources;
    }

    // ---------------------------------------------------------- Public methods

    /**
     * Preparation for the synchronization. If <i>sources</i> is not null,
     * the preparation works on the given sources. Otherwise it works on the
     * sources as they were set in the constructor.
     * Note that in the case of slow sync, the sync analysis to determine the
     * new items on the server that must be returned to the client, we need to
     * wait for the last call to this method. This is used for protocols that
     * supports a multi message session. At the contrary, client side
     * modifications can be processed immediately.
     *
     *
     * @param sources the sources to be synchronized
     * @param nextSync the timestamp of the beginning of this synchronization
     * @param principal the entity for which the synchronization is required
     * @param syncFilterType the type of the filter used
     * @param filter the filter to use in the sync
     * @param last is this the last call to prepareSlowSync ?
     *
     * @return an array of SyncOperation, one for each SyncItem that must be
     *         created/updated/deleted or in conflict.
     *
     * @see com.funambol.framework.engine.SyncStrategy
     */
    public SyncOperation[] prepareSlowSync(SyncSource[] sources,
                                           Principal principal,
                                           int syncFilterType,
                                           Timestamp nextSync,
                                           boolean last) throws SyncException {

        if (sources != null) {
            this.sources = sources;
        }

        SyncItem[] Am, Bm;
        List syncOperations = new ArrayList();

        Logger log = Logger.getLogger(LOG_NAME);

        if (log.isLoggable(Level.INFO)) {
            log.info("Preparing slow synchronization of source '"
                     + sources[1].getSourceURI()
                     + "' for "
                     + principal
                     + " ..."
                );
            if (last) {
                log.info("Last call");
            } else {
                log.info("Not the last call");
            }
        }

        //
        // First process the items provided by the client
        //
        Am = ( (MemorySyncSource)sources[1]).getUpdatedSyncItems();

        String uri = sources[0].getSourceURI();
        if (BmAll.get(uri) == null) {
            Bm = EngineHelper.createSyncItems(sources[0].getAllSyncItemKeys(),
                                              SyncItemState.NEW,
                                              sources[0],
                                              null,
                                              null);
            BmAll.put(uri, Bm);
        } else {
            Bm = (SyncItem[])BmAll.get(uri);
        }

        //
        // Just in case...
        //
        if (Am == null) {
            Am = new SyncItem[0];
        }

        if (Bm == null) {
            Bm = new SyncItem[0];
        }

        //
        // If any item from the client has not a corresponding mapping, the
        // server source must be queried for the item, since the client item
        // could be the same of an existing server item. In this case, the old
        // unmapped item is replaced in Map by the newly mapped item.
        //
        List newlyMappedItems = new ArrayList();
        fixMappedItems(newlyMappedItems, Am, sources[0], principal);

        //
        // Because it is a slow sync, items state must be reset to N
        //
        EngineHelper.setState(Am, SyncItemState.NEW);
        EngineHelper.setState(Bm, SyncItemState.NEW);

        for (int i = 0; i < Am.length; ++i) {
            if (! ( (SyncItemImpl)Am[i]).isMapped()) {
                syncOperations.add(checkSyncOperation(principal, nextSync, Am[i], null));
            }
        }

        Iterator        j       = newlyMappedItems.iterator();
        SyncItemMapping mapping = null;
        while (j.hasNext()) {
            mapping = (SyncItemMapping)j.next();

            syncOperations.add(checkSyncOperation(principal,
                                                  nextSync,
                                                  mapping.getSyncItemA(),
                                                  mapping.getSyncItemB()));
        }

        //
        // If it is the last call, we have also to consider what is in source
        // B but not in A
        //
        if (last) {
            //
            // In this case getAllSyncItems() returns the items in the datastore
            // reconstructed from the GUID-LUID mapping. In addition to these
            // items we have the ones sent in the last message (retrieved
            // calling getUpdatedSyncItems().
            //
            EngineHelper.setState(Am, SyncItemState.NEW);

            List all = new ArrayList(
                Arrays.asList( ( (MemorySyncSource)sources[1]).getAllSyncItems())
            );

            all.addAll(Arrays.asList(Am));

            for (int i = 0; i < Bm.length; ++i) {
                if (! (all.contains(Bm[i]))) {
                    syncOperations.add(checkSyncOperation(principal, nextSync, null, Bm[i]));
                }
            }
        }

        if (log.isLoggable(Level.FINEST)) {
            log.finest("operations: " + syncOperations);
        }

        if (log.isLoggable(Level.INFO)) {
            log.info("Preparation completed.");
        }

        return (SyncOperation[])syncOperations.toArray(new SyncOperationImpl[] {});
    }

    /**
     * Preparation for fast synchronization. If <i>sources</i> is not null,
     * the preparation operates on the given sources. Otherwise it works on the
     * sources as they were set in the constructor.
     * <p>
     * Refer to the architecture document for details about the algoritm applied.
     *
     * @param sources the sources to be synchronized
     * @param principal the entity for which the synchronization is required
     * @param mapping the current mapping
     * @param syncFilterType the type of the filter used
     * @param lastSync timestamp of the last synchronization
     * @param nextSync timestamp of the current synchronization
     * @param lastAnchor the last anchor to use in the mapping handling. This is
     *        required in order to check the item already sent with the same
     *         lastAnchor (Suspende and Resume).
     * @param last is this the last call to prepareFastSync ?
     *
     * @return an array of SyncOperation, one for each SyncItem that must be
     *         created/updated/deleted or in conflict.
     *
     * @see com.funambol.framework.engine.SyncStrategy
     */
    public SyncOperation[] prepareFastSync(SyncSource[]  sources,
                                           Principal     principal,
                                           ClientMapping mapping,
                                           int           syncFilterType,
                                           Timestamp     lastSync,
                                           Timestamp     nextSync,
                                           String        lastAnchor,
                                           boolean       last) throws SyncException {

        if (sources != null) {
            this.sources = sources;
        }

        // ---------------------------------------------------------------------

        List Am,    // items modified in A
             Bm,    // items modified in B
             AmBm,  // Am intersect Bm
             AAmBm, // items unmodified in A, but modified in B
             AmBBm; // items unmodified in B, but modified in A

        List syncOperations = null;

        SyncItem[] newA = null, newB = null,
                   updatedA = null, updatedB = null,
                   deletedA = null, deletedB = null;

        if (log.isLoggable(Level.INFO)) {
            log.info("Preparing fast synchronization of source '"
                     + sources[1].getSourceURI()
                     + "' for "
                     + principal
                     + " since "
                     + lastSync
                     + "..."
                );

            if (last) {
                log.info("Last call");
            } else {
                log.info("Not the last call");
            }
        }

        // ---------------------------------------------------------------------

        //
        // NOTE: simplified version - only two sources, the first one of which
        // is the client
        //
        newA     = ( (MemorySyncSource)sources[1]).getNewSyncItems();
        updatedA = ( (MemorySyncSource)sources[1]).getUpdatedSyncItems();
        deletedA = ( (MemorySyncSource)sources[1]).getDeletedSyncItems();

        if (log.isLoggable(Level.FINEST)) {
            log.finest("newA: " + Util.arrayToString(newA));
            log.finest("updatedA: " + Util.arrayToString(updatedA));
            log.finest("deletedA: " + Util.arrayToString(deletedA));
        }

        //
        // If there are some updated items not mapped, we have to handle them
        // as New item.
        // If there are some new items mapped, we have to handle them
        // as Updated item
        //
        EngineHelper.checkItemsState(newA);
        EngineHelper.checkItemsState(updatedA);

        String uri = sources[0].getSourceURI();
        if (newBAll.get(uri) == null) {

            if (log.isLoggable(Level.FINEST)) {
                log.finest("Detecting server changes...");
            }

            newB = EngineHelper.createSyncItems(
                sources[0].getNewSyncItemKeys(lastSync, nextSync),
                SyncItemState.NEW,
                sources[0],
                mapping,
                lastAnchor);

            updatedB = EngineHelper.createSyncItems(
                sources[0].getUpdatedSyncItemKeys(lastSync, nextSync),
                SyncItemState.UPDATED,
                sources[0],
                mapping,
                lastAnchor);

            deletedB = EngineHelper.createSyncItems(
                sources[0].getDeletedSyncItemKeys(lastSync, nextSync),
                SyncItemState.DELETED,
                sources[0],
                mapping,
                lastAnchor);

            if (log.isLoggable(Level.FINEST)) {
                log.finest("newB: "     + Util.arrayToString(newB));
                log.finest("updatedB: " + Util.arrayToString(updatedB));
                log.finest("deletedB: " + Util.arrayToString(deletedB));
            }


            //
            // The SyncSource should return empty arrays... but just in case...
            //
            if (newB == null) {
                newB = new SyncItemImpl[0];
            }
            if (updatedB == null) {
                updatedB = new SyncItemImpl[0];
            }
            if (deletedB == null) {
                deletedB = new SyncItemImpl[0];
            }

            if (log.isLoggable(Level.FINEST)) {
                log.finest("newB: "     + Util.arrayToString(newB));
                log.finest("updatedB: " + Util.arrayToString(updatedB));
                log.finest("deletedB: " + Util.arrayToString(deletedB));
            }

            newBAll.put    (uri, newB    );
            updatedBAll.put(uri, updatedB);
            deletedBAll.put(uri, deletedB);
        } else {
            newB     = (SyncItem[])newBAll.get(uri);
            updatedB = (SyncItem[])updatedBAll.get(uri);
            deletedB = (SyncItem[])deletedBAll.get(uri);
        }

        //
        // If any item from the client has not a corresponding mapping, the
        // server source must be queried for the item, since the client item
        // could be the same of an existing server item. In this case,  the old
        // unmapped item is replaced by the newly mapped item.
        //
        List newlyMappedItems = new ArrayList();

        fixMappedItems(newlyMappedItems, newA,     sources[0], principal);
        fixMappedItems(newlyMappedItems, updatedA, sources[0], principal);
        fixMappedItems(newlyMappedItems, deletedA, sources[0], principal);

        if (log.isLoggable(Level.FINEST)) {
            log.finest("Newly mapped items: " + newlyMappedItems);
        }

        Am = new ArrayList();
        Bm = new ArrayList();
        Am.addAll(Arrays.asList(newA));
        Am.addAll(Arrays.asList(updatedA));
        Am.addAll(Arrays.asList(deletedA));
        Bm.addAll(Arrays.asList(newB));
        Bm.addAll(Arrays.asList(updatedB));
        Bm.addAll(Arrays.asList(deletedB));

        if (log.isLoggable(Level.FINEST)) {
            log.finest("Am: " + Am);
            log.finest("Bm: " + Bm);
            log.finest("Am-Bm: " + ListUtils.subtract(Am, Bm));
            log.finest("Bm-Am: " + ListUtils.subtract(Bm, Am));
        }

        //
        // Now calculate subsets: AmBm, AmBBm, AAmBm.
        // Note that Bm and AAmBm must be calculated only when we got the
        // last message. For the others, it must be empty. At the contrary,
        //
        AmBm = EngineHelper.intersect(Am, Bm);

        //
        // We have to remove from AmBm the newlyMappedItems and then re-add them.
        // This is done because otherwise we can have the same mapping in AmBm and
        // in newlyMappedItems. However, we have to keep the mapping in the
        // newlyMappedItems because here we could have a merged item.
        //
        // Example: Conflict NEW-NEW: the server has to merge the items but in Bm
        // we have the item with the original data (before the merge operation)
        //

        //
        // The remove is based on SyncItemM apping.equals method.
        //
        AmBm.removeAll(newlyMappedItems);

        AmBm.addAll(newlyMappedItems);

        AmBBm = EngineHelper.buildAmBBm(ListUtils.subtract(Am, Bm), sources[0], principal);

        AAmBm = new ArrayList();
        if (last) {
            //
            // Note: being this class a concrete SyncStrategy implementation for
            // Funambol we know already that sources[1] is a MemorySyncSource.
            //
            newA     = ( (MemorySyncSource)sources[1]).getAllNewSyncItems();
            updatedA = ( (MemorySyncSource)sources[1]).getAllUpdatedSyncItems();
            deletedA = ( (MemorySyncSource)sources[1]).getAllDeletedSyncItems();

            SyncItemKey[] allItemsKeys = sources[0].getAllSyncItemKeys();

            if (log.isLoggable(Level.FINEST)) {
                log.finest("allItemsKeys: " + Util.arrayToString(allItemsKeys));
            }

            checkForItemsToDelete(syncFilterType, allItemsKeys, mapping, Bm);

            checkForItemsToAdd(allItemsKeys, mapping, Bm);

            AAmBm = EngineHelper.buildAAmBm(ListUtils.subtract(Bm, Am), sources[1], principal);
        } else {
            Bm.clear();
        }

        if (log.isLoggable(Level.FINEST)) {
            log.finest("AmBm:  " + AmBm);
            log.finest("AmBBm: " + AmBBm);
            log.finest("AAmBm: " + AAmBm);
        }

        //
        // Ready for conflict detection!
        //
        syncOperations =
            checkSyncOperations(principal, nextSync, Am, Bm, AmBm, AmBBm, AAmBm);

        if (log.isLoggable(Level.FINEST)) {
            log.finest("operations: " + syncOperations);
        }

        //
        // We have to remove from xxxB the items in newlyMappedItems because if
        // we have an item here, it's already handled
        //
        newB     = EngineHelper.removeItemsInOperations(newB,     syncOperations, false);
        updatedB = EngineHelper.removeItemsInOperations(updatedB, syncOperations, false);
        deletedB = EngineHelper.removeItemsInOperations(deletedB, syncOperations, false);

        newBAll.put    (uri, newB    );
        updatedBAll.put(uri, updatedB);
        deletedBAll.put(uri, deletedB);

        if (log.isLoggable(Level.INFO)) {
            log.info("Preparation completed.");
        }

        return (SyncOperation[])syncOperations.toArray(new SyncOperationImpl[] {});
    }

    /**
     * Implements Synchronizable.sync
     */
    public SyncOperationStatus[] sync(SyncOperation[] syncOperations,
                                      int syncFilterType) {
        Logger log = Logger.getLogger(LOG_NAME);

        if (log.isLoggable(Level.INFO)) {
            log.info("Synchronizing...");
        }

        if ( (syncOperations == null) || (syncOperations.length == 0)) {
            return new SyncOperationStatus[0];
        }

        List status = new ArrayList();

        SyncOperationStatus[] operationStatus = null;
        for (int i = 0; i < syncOperations.length; ++i) {
            if (log.isLoggable(Level.FINEST)) {
                log.finest("Executing " + syncOperations[i]);
            }

            //
            // execSyncOperation can return more than one status for one
            // operation when more than one source are involved
            //
            operationStatus = execSyncOperation( (SyncOperationImpl)syncOperations[i],
                                                syncFilterType);

            for (int j = 0; j < operationStatus.length; ++j) {
                status.add(operationStatus[j]);
            } // next j
        } // next i

        if (log.isLoggable(Level.FINEST)) {
            log.finest("status: " + status);
        }

        return (SyncOperationStatus[])status.toArray(new SyncOperationStatus[0]);
    }

    /**
     * Implements Synchronizable.endSync
     */
    public void endSync() throws SyncException {
        Logger log = Logger.getLogger(LOG_NAME);

        if (log.isLoggable(Level.INFO)) {
            log.info("Synchronization completed.");
        }
    }

    // -------------------------------------------------------- Protected methds

    /**
     * Checks the given SyncItem lists and creates the needed SyncOperations
     * following the rules described in the class description and in the
     * architecture document.
     *
     * @param principal who has requested the synchronization
     * @param nextSync timestamp of the current synchronization
     * @param Am the Am set
     * @param Bm the Bm set
     * @param AmBm the AmBm
     * @param AmBBm the AmBBm
     * @param AAmBm the AAmBm
     *
     * @return an List containing all the collected sync operations
     */
    protected List checkSyncOperations(Principal principal,
                                       Timestamp nextSync,
                                       List Am,
                                       List Bm,
                                       List AmBm,
                                       List AmBBm,
                                       List AAmBm) {
        SyncItemMapping mapping = null;
        SyncItem syncItemA = null, syncItemB = null;

        List all = new ArrayList();
        List operations = new ArrayList();

        // ---------------------------------------------------------------------

        all.addAll(AmBm);
        all.addAll(AmBBm);
        all.addAll(AAmBm);

        //
        // 1st check: items in both sources
        //
        Iterator i = all.iterator();
        while (i.hasNext()) {
            mapping = (SyncItemMapping)i.next();

            syncItemA = mapping.getSyncItemA();
            syncItemB = mapping.getSyncItemB();

            operations.add(
                checkSyncOperation(
                    principal,
                    nextSync,
                    syncItemA,
                    syncItemB
                )
            );
            Am.remove(syncItemA);
            Bm.remove(syncItemB);
        }

        //
        // 2nd check: items in source A and not in source B
        //
        i = Am.iterator();
        while (i.hasNext()) {
            syncItemA = (SyncItem)i.next();

            operations.add(checkSyncOperation(principal, nextSync, syncItemA, null));
        } // next i

        //
        // 3rd check: items in source B and not in source A
        //
        i = Bm.iterator();
        while (i.hasNext()) {
            syncItemB = (SyncItem)i.next();

            operations.add(checkSyncOperation(principal, nextSync, null, syncItemB));
        } // next i

        return operations;
    }

    /**
     * Create a SyncOperation based on the state of the given SyncItem couple.
     *
     * @param principal     the entity that wnats to do the operation
     * @param nextSync      timestamp of the current synchronization (used for "B" operation)
     * @param syncItemA     the SyncItem of the source A - NULL means <i>not existing</i/
     * @param syncItemB     the SyncItem of the source B - NULL means <i>not existing</i/
     *
     * @return the SyncOperation object
     */
    protected SyncOperation checkSyncOperation(Principal principal,
                                               Timestamp nextSync,
                                               SyncItem syncItemA,
                                               SyncItem syncItemB) {
        if (log.isLoggable(Level.FINEST)) {
            log.finest("check: syncItemA: "
                       + syncItemA
                       + " syncItemB: "
                       + syncItemB
            );
        }

        if (syncItemA == null) {
            syncItemA = SyncItemImpl.getNotExistingSyncItem(null);
        }
        if (syncItemB == null) {
            syncItemB = SyncItemImpl.getNotExistingSyncItem(null);
        }

        switch (syncItemA.getState()) {
            //
            // NEW
            //
            case SyncItemState.NEW:
                switch (syncItemB.getState()) {
                    case SyncItemState.NEW:
                        return new SyncConflict(principal, syncItemA, syncItemB,
                                                SyncConflict.STATE_NEW_NEW);
                    case SyncItemState.UPDATED:
                        return new SyncConflict(principal, syncItemA, syncItemB,
                                                SyncConflict.STATE_NEW_UPDATED);
                    case SyncItemState.DELETED:
                        return new SyncConflict(principal, syncItemA, syncItemB,
                                                SyncConflict.STATE_NEW_DELETED);

                    case SyncItemState.SYNCHRONIZED:
                        return new SyncConflict(principal, syncItemA, syncItemB,
                                                SyncConflict.STATE_NEW_SYNCHRONIZED);

                    case SyncItemState.NOT_EXISTING:
                        syncItemA.setTimestamp(nextSync);
                        return new SyncOperationImpl(principal, syncItemA, syncItemB,
                            SyncOperation.NEW, false, true);
                } // end inner switch

                //
                // DELETED
                //
            case SyncItemState.DELETED:
                switch (syncItemB.getState()) {
                    case SyncItemState.NEW:
                        return new SyncConflict(principal,
                                                syncItemA,
                                                syncItemB,
                                                SyncConflict.STATE_DELETED_NEW);

                    case SyncItemState.UPDATED:
                        syncItemB.setState(SyncItemState.NEW);
                        return new SyncConflict(principal,
                                                syncItemA,
                                                syncItemB,
                                                SyncConflict.STATE_DELETED_NEW);

                    case SyncItemState.SYNCHRONIZED:
                        return new SyncOperationImpl(principal,
                            syncItemA,
                            syncItemB,
                            SyncOperation.DELETE,
                            false,
                            true);

                    case SyncItemState.DELETED:
                        return new SyncConflict(principal,
                                                syncItemA,
                                                syncItemB,
                                                SyncConflict.STATE_DELETED_DELETED);

                    case SyncItemState.NOT_EXISTING:
                        return new SyncOperationImpl(principal,
                            syncItemA,
                            syncItemB,
                            SyncOperation.NOP,
                            false,
                            false);
                } // end inner switch

                //
                // UPDATED
                //
            case SyncItemState.UPDATED:
                switch (syncItemB.getState()) {
                    case SyncItemState.NEW:
                        return new SyncConflict(principal, syncItemA, syncItemB,
                                                SyncConflict.STATE_UPDATED_NEW);

                    case SyncItemState.UPDATED:
                        return new SyncConflict(principal, syncItemA, syncItemB,
                                                SyncConflict.STATE_UPDATED_UPDATED);

                    case SyncItemState.DELETED:
                        return new SyncConflict(principal, syncItemA, syncItemB,
                                                SyncConflict.STATE_UPDATED_DELETED);
                    case SyncItemState.SYNCHRONIZED:
                        syncItemA.setTimestamp(nextSync);
                        return new SyncOperationImpl(principal, syncItemA, syncItemB,
                            SyncOperation.UPDATE, false, true);
                    case SyncItemState.NOT_EXISTING:
                        syncItemA.setTimestamp(nextSync);
                        return new SyncOperationImpl(principal, syncItemA, syncItemB,
                            SyncOperation.NEW, false, true);
                } // end inner switch

                //
                // SYNCHRONIZED
                //
            case SyncItemState.SYNCHRONIZED:
                switch (syncItemB.getState()) {
                    case SyncItemState.NEW:
                        return new SyncConflict(principal, syncItemA, syncItemB,
                                                SyncConflict.STATE_SYNCHRONIZED_NEW);

                    case SyncItemState.UPDATED:
                        return new SyncOperationImpl(principal, syncItemA, syncItemB,
                            SyncOperation.UPDATE, true, false);
                    case SyncItemState.DELETED:
                        return new SyncOperationImpl(principal, syncItemA, syncItemB,
                            SyncOperation.DELETE, true, false);
                    case SyncItemState.SYNCHRONIZED:
                        return new SyncOperationImpl(principal, syncItemA, syncItemB,
                            SyncOperation.NOP, false, false);
                    case SyncItemState.NOT_EXISTING:
                        return new SyncOperationImpl(principal, syncItemA, syncItemB,
                            SyncOperation.NEW, false, true);
                } // end inner switch

                //
                // NOTEXISITNG
                //
            case SyncItemState.NOT_EXISTING:
                switch (syncItemB.getState()) {
                    case SyncItemState.NEW:
                    case SyncItemState.UPDATED:
                        return new SyncOperationImpl(principal, syncItemA, syncItemB,
                            SyncOperation.NEW, true, false);
                    case SyncItemState.SYNCHRONIZED:
                        return new SyncOperationImpl(principal, syncItemA, syncItemB,
                            SyncOperation.NEW, true, false);
                    case SyncItemState.NOT_EXISTING:
                    case SyncItemState.DELETED:
                        return new SyncOperationImpl(principal, syncItemA, syncItemB,
                            SyncOperation.NOP, false, false);
                } // end inner switch

                //
                // PARTIAL
                //
            case SyncItemState.PARTIAL:
                switch (syncItemB.getState()) {
                    case SyncItemState.NEW:
                        return new SyncConflict(principal, syncItemA, syncItemB,
                                                SyncConflict.STATE_UPDATED_NEW);

                    case SyncItemState.UPDATED:
                        return new SyncConflict(principal, syncItemA, syncItemB,
                                                SyncConflict.STATE_UPDATED_UPDATED);

                    case SyncItemState.DELETED:
                        return new SyncConflict(principal, syncItemA, syncItemB,
                                                SyncConflict.STATE_UPDATED_DELETED);

                    case SyncItemState.SYNCHRONIZED:
                        return new SyncOperationImpl(principal, syncItemA, null,
                            SyncOperation.ACCEPT_CHUNK, false, true);
                    case SyncItemState.NOT_EXISTING:
                        return new SyncOperationImpl(principal, syncItemA, null,
                            SyncOperation.ACCEPT_CHUNK, false, true);
                } // end inner switch

                //
                // CONFLICT
                // In this case itemA has a mapped twin and itemB not existing
                //
            case SyncItemState.CONFLICT:
                return new SyncConflict(principal, syncItemA, syncItemB,
                                        SyncConflict.STATE_CONFLICT_NONE);
        } // end switch

        return new SyncOperationImpl(principal, syncItemA, syncItemB, SyncOperation.NOP, false, false);
    }

    /**
     * Executes the given SyncOperation. Note that conflicts are ignored!
     * <p>
     * Note also that the number of status returned is equal to the number of
     * sources affected by the operation (0 for NOP, 1 for NEW and UPDATE and
     * 1 or 2 for DELETE).
     *
     * @param operation the SyncOperation to execute
     * @param syncFilterType the type of the filter used in the sync
     *
     * @return an array of <i>SyncOperationStatus</i> objects representing the
     *         status of the executed operation. For instance, in case of error,
     *         <i>SyncOperation.error</i> will be set to the catched exception.
     */
    protected SyncOperationStatus[] execSyncOperation(SyncOperationImpl operation,
        int syncFilterType) {

        SyncOperationStatus[] status = null;

        switch (operation.getOperation()) {
            case SyncOperation.NEW:

                status = execNewOperation(operation, syncFilterType);
                break;

            case SyncOperation.UPDATE:

                status = execUpdateOperation(operation, syncFilterType);
                break;

            case SyncOperation.DELETE:

                status = execDeleteOperation(operation, syncFilterType);
                break;

            case SyncOperation.NOP:

                status = execNOPOperation(operation, syncFilterType);
                break;

            case SyncOperation.ACCEPT_CHUNK:

                status = execAcceptChunkOperation(operation, syncFilterType);
                break;

            case SyncOperation.CONFLICT:

                status = execConflict(operation, syncFilterType);
                break;

        } // end switch

        return status;
    }

    // --------------------------------------------------------- Private methods

    /**
     * If any item from the client has not a corresponding mapping, the
     * server source must be queried for the item, since the client item
     * could be the same of an existing server item. In this case, the old
     * unmapped item is replaced in Am by the newly mapped item. Furthermore, the
     * client item and the twin are merged. The new merged data is stored in
     * itemB.
     *
     * @param newlyMappedItems the collection that will contain the newly mapped
     *        items. If null, it won't be used.
     * @param syncItems the items to be searched if not mapped
     * @param source the source that has to be queried for twins
     * @param principal the principal items are belonging to
     *
     */
    private void fixMappedItems(Collection newlyMappedItems,
                                SyncItem[] syncItems,
                                SyncSource source,
                                Principal principal) {
        SyncItem itemB = null;
        SyncItem newItemB = null;
        Map propertiesB = null;
        SyncItemKey[] twinsKey = null;
        for (int i = 0; ( (syncItems != null) && (i < syncItems.length)); ++i) {
            itemB = null;

            //
            // We have to exclude the item already mapped, the item not complete
            // and the item deleted (this items don't contain data)
            //
            if ( (! ( (SyncItemImpl)syncItems[i]).isMapped()) &&
                (SyncItemState.PARTIAL != syncItems[i].getState()) &&
                (SyncItemState.DELETED != syncItems[i].getState())) {

                try {
                    twinsKey = source.getSyncItemKeysFromTwin(syncItems[i]);
                    if (twinsKey != null && twinsKey.length > 0) {
                        itemB = new SyncItemImpl(source,
                                                 twinsKey[0].getKeyValue());
                    }
                } catch (SyncSourceException e) {
                    String msg = "Error retrieving the twin item of "
                                 + syncItems[i].getKey()
                                 + " from source "
                                 + source
                                 + ": "
                                 + e.getMessage();
                    log.severe(msg);
                    log.throwing(getClass().getName(), "fixSyncMapping", e);
                }
                if (itemB != null) {
                    if (log.isLoggable(Level.FINEST)) {
                        log.finest("Hey, client item "
                                   + syncItems[i].getKey().getKeyAsString()
                                   + " is the same of server item "
                                   + itemB.getKey().getKeyAsString()
                                   + "! We have to merge the items !!"
                            );
                    }

                    itemB.setTimestamp(
                        syncItems[i].getTimestamp()
                    ); // this contains the time of the current sync

                    itemB.setState(SyncItemState.SYNCHRONIZED);

                    try {

                        //
                        // We keep the state of the clientItem because the ss
                        // can change it (but we don't want it!!!)
                        //
                        char state = syncItems[i].getState();

                        boolean itemChanged
                            = source.mergeSyncItems(itemB.getKey(), syncItems[i]);

                        //
                        // We restore the state of the item
                        //
                        syncItems[i].setState(state);

                        // propertiesB = itemB.getProperties();

                        newItemB = new SyncItemImpl(
                            itemB.getSyncSource(),
                            itemB.getKey().getKeyAsString(),
                            syncItems[i].getKey().getKeyAsString(),
                            syncItems[i].getState(),
                            syncItems[i].getContent(),
                            syncItems[i].getFormat(),
                            syncItems[i].getType(),
                            syncItems[i].getTimestamp()
                        );

                        // newItemB.setProperties(propertiesB);

                        String s = new String(syncItems[i].getContent());

                        if (itemChanged) {
                            //
                            // We get the merge result from syncItems[i]
                            //
                            newItemB.setState(SyncItemState.UPDATED);
                        } else {
                            newItemB.setState(SyncItemState.SYNCHRONIZED);
                        }

                    } catch (SyncSourceException e) {
                        String msg = "Error merging the client item "
                                     + syncItems[i].getKey()
                                     + " with server item "
                                     + itemB.getKey().getKeyAsString()
                                     + " from source "
                                     + source
                                     + ": "
                                     + e.getMessage();
                        log.severe(msg);
                        log.throwing(getClass().getName(), "fixSyncMapping", e);
                    }

                    SyncItemMapping mapping = new SyncItemMapping(newItemB.getKey());
                    syncItems[i] = SyncItemHelper.newMappedSyncItem(newItemB.getKey(), syncItems[i]);

                    mapping.setMapping(syncItems[i], newItemB);
                    if (newlyMappedItems != null) {
                        newlyMappedItems.add(mapping);
                    }
                }
            }
        } // next i
    }

    /**
     * Verifies that the given item the size of its content metches the
     * specified size, if specified. In the case of a mismatch, a
     * <code>SyncSourceException</code> is thrown with status code
     * OBJECT_SIZE_MISMATCH (424).
     *
     * @param item item to check
     *
     * @throws ObjectSizeMismatchException if the content size does not match
               the declared size (if declared)
     */
    private void checkSize(SyncItem item) throws ObjectSizeMismatchException {

        if (log.isLoggable(Level.FINEST)) {
            log.finest("Check item size");
        }

        Long size = (Long)((SyncItemImpl)item).getPropertyValue(SyncItemImpl.PROPERTY_DECLARED_SIZE);

        if (size == null) {
            if (log.isLoggable(Level.FINEST)) {
                log.finest("Declared size is null...skip the check");
            }

            return;
        }

        if (log.isLoggable(Level.FINEST)) {
            log.finest("Declared size: " + size);
        }

        Long contentSize =
            (Long)((SyncItemImpl)item).getPropertyValue(SyncItemImpl.PROPERTY_CONTENT_SIZE);

        if (contentSize == null) {
            contentSize = new Long(item.getContent().length);
        }

        if (log.isLoggable(Level.FINEST)) {
            log.finest("Content size: " + contentSize);
        }


        //
        // We cannot check the content lenght (content.lenght) because the items
        // are already handled by the DataTransformationManager in the Sync4jEngine
        // See Sync4jEngine.applyDataTransformationOnIngoingItems
        //
        if (contentSize == null || !size.equals(contentSize)) {

            if (log.isLoggable(Level.FINEST)) {
                log.finest("The size of the received object does not match the given size (" +
                          contentSize + " vs " + size + ")");
            }

            throw new ObjectSizeMismatchException(
                "The size of the received object does not match the given size");
        }
    }


    /**
     * Executes a conflict and returns the given status
     * @param operation SyncOperationImpl
     * @param syncFilterType int
     * @return SyncOperationStatus
     * @throws SyncException
     */
    private SyncOperationStatus[] execConflict(SyncOperationImpl operation,
                                               int syncFilterType) {

        SyncItemImpl syncItemA = (SyncItemImpl)operation.getSyncItemA();

        ModificationCommand cmd =
            (ModificationCommand)syncItemA.getPropertyValue(SyncItemImpl.PROPERTY_COMMAND);

        SyncOperationStatus[] status = null;

        try {
            SyncConflict sc = (SyncConflict)operation;

            //
            //Type CX means that itemA is in state CONFLICT and itemB is in state NOT_EXISTING
            //This situation happens when itemA has already a twin mapped
            //
            if (SyncConflict.STATE_CONFLICT_NONE.equals(sc.getType())) {

                status = execConflictCX(operation, syncFilterType);

            } else if (SyncConflict.STATE_DELETED_DELETED.equals(sc.getType())) {

                status = execConflictDD(operation, syncFilterType);

            } else if (SyncConflict.STATE_NEW_NEW.equals(sc.getType()) ||
                       SyncConflict.STATE_NEW_UPDATED.equals(sc.getType())) {

                status = execConflictNU(operation, syncFilterType);

            } else if (SyncConflict.STATE_UPDATED_UPDATED.equals(sc.getType()) ||
                       SyncConflict.STATE_UPDATED_NEW.equals(sc.getType())) {

                status = execConflictUU(operation, syncFilterType);

            } else if (SyncConflict.STATE_UPDATED_DELETED.equals(sc.getType())) {

                status = execConflictUD(operation, syncFilterType);

            } else if (SyncConflict.STATE_NEW_SYNCHRONIZED.equals(sc.getType())) {

                status = execConflictNS(operation, syncFilterType);

            } else if (SyncConflict.STATE_DELETED_NEW.equals(sc.getType()) ||
                       SyncConflict.STATE_DELETED_UPDATED.equals(sc.getType())) {

                status = execConflictDU(operation, syncFilterType);

            } else {

                status = execConflictNotSpecified(operation, syncFilterType);

            }

        } catch (SyncException e) {
            log.severe("Error executing sync operation: " + e.getMessage());
            log.throwing(getClass().getName(), "execSyncOperation", e);
            status[0] = new Sync4jOperationStatusError(operation, sources[1], cmd, e);
            operation.setAOperation(false);
            operation.setBOperation(false);
        }


        return status;
    }


    /**
     * Executes a conflict DU and returns the given status
     * @param operation SyncOperationImpl
     * @param syncFilterType int
     * @return SyncOperationStatus
     * @throws SyncException
     */
    private SyncOperationStatus[] execConflictDU(SyncOperationImpl operation,
                                                 int syncFilterType)
    throws SyncException {

        SyncItemImpl syncItemA = (SyncItemImpl)operation.getSyncItemA();
        SyncItem     syncItemB = operation.getSyncItemB();

        ModificationCommand cmd =
            (ModificationCommand)syncItemA.getPropertyValue(SyncItemImpl.PROPERTY_COMMAND);

        SyncOperationStatus[] status = new SyncOperationStatus[1];

        SyncItemKey luid = ( (SyncItemImpl)syncItemA).getMappedKey();
        if (luid == null) {
            throw new SyncException("Item A with key "
                                    + syncItemA.getKey()
                                    + " should have a mapping, but the mapped key is null"
            );
        }

        boolean isItemInFilter = true;

        if (syncFilterType != SYNC_WITHOUT_FILTER) {
            //
            // We have to use the key because we don't have the content
            //
            isItemInFilter = sources[0].isSyncItemInFilterClause(syncItemB.getKey());

            if (!isItemInFilter) {

                status[0] = new Sync4jOperationStatusConflict(
                    operation,
                    sources[1],
                    cmd,
                    StatusCode.ITEM_NOT_DELETED
                );

                operation.setAOperation(false);
                operation.setBOperation(false);

                return status;

            }
        }

        //
        // In all others cases we re-add the item on the client
        // (with status code 419)
        //
        SyncItemImpl resolvingItem = null;

        //
        // We re-add the item on the client
        //
        resolvingItem = new SyncItemImpl(
            syncItemB.getSyncSource(),
            syncItemA.getMappedKey().getKeyAsString(),
            null,
            SyncItemState.UNKNOWN,
            syncItemB.getContent(),
            syncItemB.getFormat(),
            syncItemB.getType(),
            syncItemB.getTimestamp()
        );

        if (resolvingItem.isMapped()) {
            syncItemA = (SyncItemImpl) (sources[1].updateSyncItem(resolvingItem));
        } else {
            syncItemA = (SyncItemImpl) (sources[1].addSyncItem(resolvingItem));
        }

        operation.setSyncItemA(syncItemA);
        operation.setAOperation(true);

        status[0] = new Sync4jOperationStatusConflict(
            operation,
            sources[1],
            cmd,
            StatusCode.CONFLICT_RESOLVED_WITH_SERVER_DATA
        );

        return status;
    }

    /**
     * Executes a conflict DD and returns the given status
     * @param operation SyncOperationImpl
     * @param syncFilterType int
     * @return SyncOperationStatus
     * @throws SyncException
     */
    private SyncOperationStatus[] execConflictDD(SyncOperationImpl operation,
                                                 int syncFilterType)
    throws SyncException {

        SyncItemImpl syncItemA = (SyncItemImpl)operation.getSyncItemA();

        ModificationCommand cmd =
            (ModificationCommand)syncItemA.getPropertyValue(SyncItemImpl.PROPERTY_COMMAND);

        SyncOperationStatus[] status = new SyncOperationStatus[1];

        status[0] = new Sync4jOperationStatusConflict(
            operation,
            sources[1],
            cmd,
            StatusCode.ITEM_NOT_DELETED
        );

        return status;
    }

    /**
     * Executes a conflict CX and returns the given status
     * @param operation SyncOperationImpl
     * @param syncFilterType int
     * @return SyncOperationStatus
     * @throws SyncException
     */
    private SyncOperationStatus[] execConflictCX(SyncOperationImpl operation,
                                                 int syncFilterType)
    throws SyncException {

        SyncItemImpl syncItemA = (SyncItemImpl)operation.getSyncItemA();

        ModificationCommand cmd =
            (ModificationCommand)syncItemA.getPropertyValue(SyncItemImpl.PROPERTY_COMMAND);

        SyncOperationStatus[] status = new SyncOperationStatus[1];

        sources[1].removeSyncItem(syncItemA.getMappedKey(),
                                  syncItemA.getTimestamp());

        status[0] = new Sync4jOperationStatusConflict(
            operation,
            sources[1],
            cmd,
            StatusCode.CONFLICT_RESOLVED_WITH_SERVER_DATA
        );

        return status;
    }

    /**
     * Executes a conflict NS and returns the given status
     * @param operation SyncOperationImpl
     * @param syncFilterType int
     * @return SyncOperationStatus
     * @throws SyncException
     */
    private SyncOperationStatus[] execConflictNS(SyncOperationImpl operation,
                                                 int syncFilterType)
    throws SyncException {

        SyncItemImpl syncItemA = (SyncItemImpl)operation.getSyncItemA();
        SyncItem     syncItemB = operation.getSyncItemB();

        ModificationCommand cmd =
            (ModificationCommand)syncItemA.getPropertyValue(SyncItemImpl.PROPERTY_COMMAND);

        SyncOperationStatus[] status = new SyncOperationStatus[1];

        //
        // This happens when the Client sent an item that have a twin
        // synchronized on server.
        //

        SyncItemKey luid = ( (SyncItemImpl)syncItemA).getMappedKey();
        if (luid == null) {
            throw new SyncException("Item A with key "
                + syncItemA.getKey()
                + " should have a mapping, but the mapped key is null"
                );
        }

        if (syncFilterType == SYNC_WITH_EXCLUSIVE_FILTER &&
            !sources[0].isSyncItemInFilterClause(syncItemA)) {

            sources[1].removeSyncItem(luid, syncItemB.getTimestamp());

            operation.setDeleteForced(true);
            operation.setAOperation(true);

        } else {

            operation.setAOperation(false);

        }

        status[0] = new Sync4jOperationStatusConflict(
            operation,
            sources[1],
            cmd,
            StatusCode.ITEM_ADDED
        );

        return status;
    }

    /**
     * Executes a conflict not handle by the others cases
     * (this is the default conflict handling)
     * @param operation SyncOperationImpl
     * @param syncFilterType int
     * @return SyncOperationStatus
     * @throws SyncException
     */
    private SyncOperationStatus[] execConflictNotSpecified(SyncOperationImpl operation,
                                                           int syncFilterType)
    throws SyncException {

        SyncItemImpl syncItemA = (SyncItemImpl)operation.getSyncItemA();
        SyncItem     syncItemB = operation.getSyncItemB();

        ModificationCommand cmd =
            (ModificationCommand)syncItemA.getPropertyValue(SyncItemImpl.PROPERTY_COMMAND);

        SyncOperationStatus[] status = new SyncOperationStatus[1];

        SyncItemImpl resolvingItem = null;

        SyncItemKey luid = ( (SyncItemImpl)syncItemA).getMappedKey();
        if (luid == null) {
            throw new SyncException("Item A with key "
                + syncItemA.getKey()
                + " should have a mapping, but the mapped key is null"
                );
        }

        if (syncFilterType == SYNC_WITH_EXCLUSIVE_FILTER &&
            !sources[0].isSyncItemInFilterClause(syncItemA)) {

            sources[1].removeSyncItem(luid, syncItemB.getTimestamp());

            operation.setDeleteForced(true);
            operation.setAOperation(true);

        } else {

            resolvingItem = new SyncItemImpl(
                syncItemB.getSyncSource(),
                syncItemA.getMappedKey().getKeyAsString(),
                null,
                SyncItemState.UNKNOWN,
                syncItemB.getContent(),
                syncItemB.getFormat(),
                syncItemB.getType(),
                syncItemB.getTimestamp()
            );

            if (resolvingItem.isMapped()) {
                syncItemA = (SyncItemImpl) (sources[1].updateSyncItem(resolvingItem));
            } else {
                syncItemA = (SyncItemImpl) (sources[1].addSyncItem(resolvingItem));
            }

            operation.setSyncItemA(syncItemA);
            operation.setAOperation(true);

        }

        status[0] = new Sync4jOperationStatusConflict(
            operation,
            sources[1],
            cmd,
            StatusCode.CONFLICT_RESOLVED_WITH_SERVER_DATA
        );

        return status;
    }


    /**
     * Executes a conflict NU and returns the given status
     * @param operation SyncOperationImpl
     * @param syncFilterType int
     * @return SyncOperationStatus
     * @throws SyncException
     */
    private SyncOperationStatus[] execConflictNU(SyncOperationImpl operation,
                                                 int syncFilterType)
    throws SyncException {

        SyncItemImpl syncItemA = (SyncItemImpl)operation.getSyncItemA();
        SyncItem     syncItemB = operation.getSyncItemB();

        ModificationCommand cmd =
            (ModificationCommand)syncItemA.getPropertyValue(SyncItemImpl.PROPERTY_COMMAND);

        SyncOperationStatus[] status = new SyncOperationStatus[1];

        SyncItemImpl resolvingItem = null;

        //
        // In a conflict raised by a twin item the mappedKey
        // is not null. There are these conflicts even when
        // the Engine maps two items with the same key (the key
        // on the client is the same key on the server). In
        // this case the mappedKey is null, so we have to used
        // SyncItem.getKey();
        //
        SyncItemKey mappedKey = null;
        if ( ( (SyncItemImpl)syncItemA).getMappedKey() == null) {
            mappedKey = syncItemA.getKey();
        } else {
            mappedKey = syncItemA.getMappedKey();
        }

        boolean isItemInFilter = true;

        if (syncFilterType != SYNC_WITHOUT_FILTER) {
            isItemInFilter = sources[0].isSyncItemInFilterClause(syncItemA);
        }

        if (syncFilterType == SYNC_WITH_EXCLUSIVE_FILTER) {
            //
            // exclusive filter
            //
            if (!isItemInFilter) {
                operation.setDeleteForced(true);
                sources[1].removeSyncItem(mappedKey, syncItemB.getTimestamp());
            }

        } else if (syncFilterType == SYNC_WITH_INCLUSIVE_FILTER) {
            //
            // inclusive filter:
            // if the item isn't in the filter criteria
            // we convert the operation in NOP and  return a 200 status code
            // Otherwise a CONFLICT_RESOLVED_WITH_MERGE is returned
            //
            if (!isItemInFilter) {
                operation.setOperation(SyncOperation.NOP);

                status[0] = new Sync4jOperationStatusConflict(
                    operation,
                    sources[1],
                    cmd,
                    StatusCode.OK
                );

            } else {
                operation.setAOperation(true);

                status[0] = new Sync4jOperationStatusConflict(
                    operation,
                    sources[1],
                    cmd,
                    StatusCode.CONFLICT_RESOLVED_WITH_MERGE
                );
            }

            return status;

        } else {

            resolvingItem = new SyncItemImpl(
                syncItemB.getSyncSource(),
                syncItemA.getKey().getKeyAsString(),
                mappedKey.getKeyAsString(),
                SyncItemState.UNKNOWN,
                syncItemB.getContent(),
                syncItemB.getFormat(),
                syncItemB.getType(),
                syncItemB.getTimestamp()
            );

            if (resolvingItem.isMapped()) {
                syncItemA = (SyncItemImpl) (sources[1].updateSyncItem(resolvingItem));
            } else {
                syncItemA = (SyncItemImpl) (sources[1].addSyncItem(resolvingItem));
            }

            operation.setSyncItemA(syncItemA);

        }

        operation.setAOperation(true);

        //
        // The conflict is resolved merging data, but the merge is
        // done into fixMappedItems methods because the NEW item
        // from the server isn't mapped and so, the methods
        // getSyncItemTwin and mergeSyncItem are called into
        // fixMappedItems
        //

        status[0] = new Sync4jOperationStatusConflict(
            operation,
            sources[1],
            cmd,
            StatusCode.CONFLICT_RESOLVED_WITH_MERGE
        );

        return status;
    }

    /**
     * Executes a conflict UU and returns the given status
     * @param operation SyncOperationImpl
     * @param syncFilterType int
     * @return SyncOperationStatus
     * @throws SyncException
     */
    private SyncOperationStatus[] execConflictUU(SyncOperationImpl operation,
                                                 int syncFilterType)
    throws SyncException {

        SyncItemImpl syncItemA = (SyncItemImpl)operation.getSyncItemA();
        SyncItem     syncItemB = operation.getSyncItemB();

        ModificationCommand cmd =
            (ModificationCommand)syncItemA.getPropertyValue(SyncItemImpl.PROPERTY_COMMAND);

        SyncOperationStatus[] status = new SyncOperationStatus[1];

        SyncItemImpl resolvingItem = null;

        SyncItemKey luid = syncItemA.getMappedKey();
        if (luid == null) {
            throw new SyncException("Item A with key "
                + syncItemA.getKey()
                + " should have a mapping, but the mapped key is null"
                );
        }

        resolvingItem = new SyncItemImpl(
            syncItemB.getSyncSource(),
            ( (SyncItemImpl)syncItemA).getMappedKey().getKeyAsString(),
            null,
            SyncItemState.UNKNOWN,
            syncItemB.getContent(),
            syncItemB.getFormat(),
            syncItemB.getType(),
            syncItemB.getTimestamp()
        );

        //
        // We have also to merge the client item with the server item
        //
        boolean itemChanged =
            sources[0].mergeSyncItems(syncItemB.getKey(),
            syncItemA);

        if (itemChanged) {

            boolean isItemInFilter = true;

            if (syncFilterType != SYNC_WITHOUT_FILTER) {
                isItemInFilter = sources[0].isSyncItemInFilterClause(syncItemB.getKey());
            }

            if (isItemInFilter) {
                resolvingItem.setContent(syncItemA.getContent());
                resolvingItem.setState(SyncItemState.UPDATED);

                if (resolvingItem.isMapped()) {
                    syncItemA = (SyncItemImpl) (sources[1].updateSyncItem(resolvingItem));
                } else {
                    syncItemA = (SyncItemImpl) (sources[1].addSyncItem(resolvingItem));
                }

                operation.setSyncItemA(syncItemA);
                operation.setAOperation(true);

                status[0] = new Sync4jOperationStatusConflict(
                    operation,
                    sources[1],
                    cmd,
                    StatusCode.CONFLICT_RESOLVED_WITH_MERGE
                );
            } else {
                if (syncFilterType == SYNC_WITH_EXCLUSIVE_FILTER) {
                    //
                    // The merged item is out the filter so the server has to delete
                    // it on the client
                    //
                    operation.setDeleteForced(true);
                    status[0] = new Sync4jOperationStatusConflict(
                        operation,
                        sources[1],
                        cmd,
                        StatusCode.CONFLICT_RESOLVED_WITH_MERGE
                    );
                } else if (syncFilterType == SYNC_WITH_INCLUSIVE_FILTER) {
                    //
                    // The merged item is out the filter so the server mustn't
                    // send it to the client
                    //
                    operation.setAOperation(false);
                    status[0] = new Sync4jOperationStatusConflict(
                        operation,
                        sources[1],
                        cmd,
                        StatusCode.CONFLICT_RESOLVED_WITH_MERGE
                    );
                }
            }
        } else {
            resolvingItem.setState(SyncItemState.SYNCHRONIZED);
            status[0] = new Sync4jOperationStatusConflict(
                operation,
                sources[1],
                cmd,
                StatusCode.OK
            );
        }

        return status;
    }

    /**
     * Executes a conflict UD and returns the given status
     * @param operation SyncOperationImpl
     * @param syncFilterType int
     * @return SyncOperationStatus
     * @throws SyncException
     */
    private SyncOperationStatus[] execConflictUD(SyncOperationImpl operation,
                                                 int syncFilterType)
    throws SyncException {

        SyncItemImpl syncItemA = (SyncItemImpl)operation.getSyncItemA();
        SyncItem     syncItemB = operation.getSyncItemB();

        ModificationCommand cmd =
            (ModificationCommand)syncItemA.getPropertyValue(SyncItemImpl.PROPERTY_COMMAND);

        SyncOperationStatus[] status = new SyncOperationStatus[1];

        SyncItemKey luid = syncItemA.getMappedKey();
        if (luid == null) {
            throw new SyncException( "Item A with key "
                                   + syncItemA.getKey()
                                   + " should have a mapping, but the mapped key is null"
                                   );
        }

        //
        // In this case the item on the server has been deleted
        // so we have to call the addSyncItem
        //
        syncItemB = sources[0].addSyncItem(syncItemA);
        operation.setSyncItemB(syncItemB);
        operation.setBOperation(true);

        status[0] = new Sync4jOperationStatusConflict(
            operation,
            sources[1],
            cmd,
            StatusCode.OK
        );

        return status;
    }

    /**
     * Executes an ACCEPT_CHUNK operation and returns the given status
     * @param operation SyncOperationImpl
     * @param syncFilterType int
     * @return SyncOperationStatus
     * @throws SyncException
     */
    private SyncOperationStatus[] execAcceptChunkOperation(SyncOperationImpl operation,
                                                            int syncFilterType) {

        SyncItemImpl syncItemA = (SyncItemImpl)operation.getSyncItemA();

        ModificationCommand cmd =
            (ModificationCommand)syncItemA.getPropertyValue(SyncItemImpl.PROPERTY_COMMAND);

        SyncOperationStatus[] status = new SyncOperationStatus[1];

        cmd = (ModificationCommand)syncItemA.getPropertyValue(SyncItemImpl.PROPERTY_COMMAND);

        status[0] = new Sync4jOperationStatusOK(
            operation, sources[0], cmd, StatusCode.CHUNKED_ITEM_ACCEPTED
        );

        return status;
    }


    /**
     * Executes a NEW operation and returns the given status
     * @param operation SyncOperationImpl
     * @param syncFilterType int
     * @return SyncOperationStatus
     */
    private SyncOperationStatus[] execNewOperation(SyncOperationImpl operation,
                                                   int syncFilterType) {

        SyncItemImpl syncItemA = (SyncItemImpl)operation.getSyncItemA();
        SyncItem     syncItemB = operation.getSyncItemB();

        ModificationCommand cmd =
            (ModificationCommand)syncItemA.getPropertyValue(SyncItemImpl.PROPERTY_COMMAND);

        SyncOperationStatus[] status = new SyncOperationStatus[1];

        status = new SyncOperationStatus[1];
        if (operation.isAOperation()) {
            cmd = (ModificationCommand)syncItemA.getPropertyValue(
                SyncItemImpl.PROPERTY_COMMAND
            );
            try {

                if ( ( (SyncItemImpl)syncItemB).isMapped()) {
                    syncItemA = (SyncItemImpl)sources[1].updateSyncItem(syncItemB);
                } else {
                    syncItemA = (SyncItemImpl)sources[1].addSyncItem(syncItemB);
                }

                operation.setSyncItemA(syncItemA);

                status[0] = new Sync4jOperationStatusOK(operation,
                    sources[1],
                    cmd,
                    StatusCode.ITEM_ADDED);

            } catch (SyncException e) {
                log.severe("Error executing sync operation: " + e.getMessage());
                log.throwing(getClass().getName(), "execSyncOperation", e);
                status[0] = new Sync4jOperationStatusError(operation, sources[1], cmd, e);
                operation.setAOperation(false);
                operation.setBOperation(false);
            }
        } else if (operation.isBOperation()) {
            cmd = (ModificationCommand)syncItemA.getPropertyValue(
                SyncItemImpl.PROPERTY_COMMAND
            );
            syncItemB.setTimestamp(syncItemA.getTimestamp()); // this contains the
                                                              // current sync
            try {
                checkSize(syncItemA);

                if (syncItemA.isMapped()) {
                    syncItemB = sources[0].updateSyncItem(syncItemA);
                } else {
                    syncItemB = sources[0].addSyncItem(syncItemA);
                }

                operation.setSyncItemB(syncItemB);

                if (syncFilterType == SYNC_WITH_EXCLUSIVE_FILTER) {
                    if (!sources[0].isSyncItemInFilterClause(syncItemB)) {
                        operation.setDeleteForced(true);
                    }
                }

                if (status[0] == null) {
                    status[0] = new Sync4jOperationStatusOK(operation,
                        sources[1],
                        cmd,
                        StatusCode.ITEM_ADDED);
                }
            } catch (ObjectSizeMismatchException e) {
                log.info(e.getMessage());
                status[0] = new Sync4jOperationStatusError(operation, sources[0], cmd, e);
                operation.setAOperation(false);
                operation.setBOperation(false);
            } catch (SyncException e) {
                log.severe("Error executing sync operation: " + e.getMessage());
                log.throwing(getClass().getName(), "execSyncOperation", e);
                status[0] = new Sync4jOperationStatusError(operation, sources[0], cmd, e);
                operation.setAOperation(false);
                operation.setBOperation(false);
            }
        }

        return status;
    }

    /**
     * Executes an UPDATE operation and returns the given status
     * @param operation SyncOperationImpl
     * @param syncFilterType int
     * @return SyncOperationStatus
     * @throws SyncException
     */
    private SyncOperationStatus[] execUpdateOperation(SyncOperationImpl operation,
                                                      int syncFilterType) {

        SyncItemImpl syncItemA = (SyncItemImpl)operation.getSyncItemA();
        SyncItem     syncItemB = operation.getSyncItemB();

        ModificationCommand cmd =
            (ModificationCommand)syncItemA.getPropertyValue(SyncItemImpl.PROPERTY_COMMAND);

        SyncOperationStatus[] status = new SyncOperationStatus[1];
        if (operation.isAOperation()) {
            try {

                if ( ( (SyncItemImpl)syncItemB).isMapped()) {
                    syncItemA = (SyncItemImpl) (sources[1].updateSyncItem(syncItemB));
                } else {
                    syncItemA = (SyncItemImpl) (sources[1].addSyncItem(syncItemB));
                }

                operation.setSyncItemA(syncItemA);
                status[0] = new Sync4jOperationStatusOK(operation, sources[1], cmd);
            } catch (SyncException e) {
                log.severe("Error executing sync operation: " + e.getMessage());
                log.throwing(getClass().getName(), "execSyncOperation", e);
                status[0] = new Sync4jOperationStatusError(operation, sources[1], cmd, e);
                operation.setAOperation(false);
                operation.setBOperation(false);
            }
        } else if (operation.isBOperation()) {

            syncItemB.setTimestamp(syncItemA.getTimestamp()); // this contains the
                                                              // current sync
            try {
                checkSize(syncItemA);

                if ( ( (SyncItemImpl)syncItemA).isMapped()) {
                    syncItemB = sources[0].updateSyncItem(syncItemA);
                } else {
                    syncItemB = sources[0].addSyncItem(syncItemA);
                }

                operation.setSyncItemB(syncItemB);

                if (syncFilterType == SYNC_WITH_EXCLUSIVE_FILTER) {
                    if (!sources[0].isSyncItemInFilterClause(syncItemB)) {
                        operation.setDeleteForced(true);
                    }
                }
                if (status[0] == null) {
                    status[0] = new Sync4jOperationStatusOK(operation, sources[0], cmd);
                }
            } catch (ObjectSizeMismatchException e) {
                log.info(e.getMessage());
                status[0] = new Sync4jOperationStatusError(operation, sources[0], cmd, e);
                operation.setAOperation(false);
                operation.setBOperation(false);
            } catch (SyncException e) {
                log.severe("Error executing sync operation: " + e.getMessage());
                log.throwing(getClass().getName(), "execSyncOperation", e);
                status[0] = new Sync4jOperationStatusError(operation, sources[0], cmd, e);
                operation.setAOperation(false);
                operation.setBOperation(false);
            }
        }
        return status;
    }


    /**
     * Executes a DELETE operation and returns the given status
     * @param operation SyncOperationImpl
     * @param syncFilterType int
     * @return SyncOperationStatus
     * @throws SyncException
     */
    private SyncOperationStatus[] execDeleteOperation(SyncOperationImpl operation,
                                                      int syncFilterType) {

        SyncItemImpl syncItemA = (SyncItemImpl)operation.getSyncItemA();
        SyncItem     syncItemB = operation.getSyncItemB();

        ModificationCommand cmd =
            (ModificationCommand)syncItemA.getPropertyValue(SyncItemImpl.PROPERTY_COMMAND);

        SyncOperationStatus[] status = null;

        //
        // How many status we need? One if we have to delete the item
        // from only one source, two if we have to delete it from both
        // sources
        //
        int size = 1;
        if (operation.isAOperation() && operation.isBOperation()) {
            size = 2;
        }
        status = new SyncOperationStatus[size];

        int s = 0;
        if (operation.isBOperation()) {
            syncItemB.setTimestamp(syncItemA.getTimestamp()); // this contains the
                                                              // timestamp of the
                                                              // current sync

            boolean softDelete = false;

            if (cmd instanceof Delete) {
                softDelete = ( (Delete)cmd).isSftDel();
            }
            try {
                sources[0].removeSyncItem(syncItemB.getKey(),
                                          syncItemB.getTimestamp(),
                                          softDelete);
                status[s++] = new Sync4jOperationStatusOK(operation, sources[0], cmd);
            } catch (SyncException e) {
                log.severe("Error executing sync operation: " + e.getMessage());
                log.throwing(getClass().getName(), "execSyncOperation", e);
                status[s++] = new Sync4jOperationStatusError(operation, sources[0], cmd, e);
                operation.setAOperation(false);
                operation.setBOperation(false);
            }
        }
        if (operation.isAOperation()) {

            try {
                sources[1].removeSyncItem(syncItemA.getKey(),
                                          syncItemA.getTimestamp());

                status[s++] = new Sync4jOperationStatusOK(operation, sources[1], cmd);
            } catch (SyncException e) {
                log.severe("Error executing sync operation: " + e.getMessage());
                log.throwing(getClass().getName(), "execSyncOperation", e);
                status[s++] = new Sync4jOperationStatusError(operation, sources[1], cmd, e);
                operation.setAOperation(false);
                operation.setBOperation(false);
            }
        }

        return status;
    }


    /**
     * Executes a NOP operation and returns the given status
     * @param operation SyncOperationImpl
     * @param syncFilterType int
     * @return SyncOperationStatus
     * @throws SyncException
     */
    private SyncOperationStatus[] execNOPOperation(SyncOperationImpl operation,
                                                   int syncFilterType) {

        SyncItemImpl syncItemA = (SyncItemImpl)operation.getSyncItemA();
        SyncItem     syncItemB = operation.getSyncItemB();

        ModificationCommand cmd =
            (ModificationCommand)syncItemA.getPropertyValue(SyncItemImpl.PROPERTY_COMMAND);

        SyncOperationStatus[] status = null;

        //
        // A NOP operation can be due by one of the following conditions
        // (see checkOperation(...)):
        //
        // 1. both items are flagged as "Deleted"
        // 2. both items are flagged "Synchronized"
        // 3. itemA is "Deleted" and itemB is "Not existing" or "Deleted"
        //
        // In case 1. we want a status is returned, thought no real
        // action is performed.
        //
        int s    = 0;
        int size = 0;

        if (operation.isAOperation())++size;
        if (operation.isBOperation())++size;

        status = new SyncOperationStatus[size];

        if (operation.isBOperation()) {
            status[s++] = new Sync4jOperationStatusOK(operation, sources[0], cmd);
        }

        if (operation.isAOperation()) {
            status[s++] = new Sync4jOperationStatusOK(operation, sources[1], cmd);
        }

        if (!operation.isAOperation() && !operation.isBOperation()) {
            char stateItemA = syncItemA.getState();
            char stateItemB = syncItemB.getState();

            switch (stateItemA) {
                case SyncItemState.SYNCHRONIZED:
                    if (stateItemB == SyncItemState.SYNCHRONIZED) {
                        status = new SyncOperationStatus[1];

                        if (cmd instanceof Add) {
                            status[0] = new Sync4jOperationStatusOK(operation, sources[1],
                                cmd, StatusCode.ALREADY_EXISTS);
                        } else {
                            status[0] = new Sync4jOperationStatusOK(operation, sources[1],
                                cmd);
                        }

                        try {
                            if (syncFilterType == SYNC_WITH_EXCLUSIVE_FILTER) {
                                if (!sources[0].isSyncItemInFilterClause(syncItemA)) {
                                    operation.setDeleteForced(true);
                                }
                            }
                        } catch (SyncSourceException e) {
                            log.severe("Error executing sync operation: " + e.getMessage());
                            log.throwing(getClass().getName(), "execSyncOperation", e);
                            status[0] = new Sync4jOperationStatusError(operation, sources[0],
                                cmd, e);
                            operation.setAOperation(false);
                            operation.setBOperation(false);
                        }
                        break ;
                    }
                case SyncItemState.DELETED:
                    if (stateItemB == SyncItemState.NOT_EXISTING) {
                        //
                        //client sent Delete item that not exist on server
                        //
                        operation.setSyncItemA(syncItemA);

                        status = new SyncOperationStatus[1];
                        status[0] = new Sync4jOperationStatusOK(
                            operation,
                            sources[1],
                            cmd,
                            StatusCode.ITEM_NOT_DELETED
                                    );
                        break;

                    } else if (stateItemB == SyncItemState.DELETED) {
                        //
                        //client sent Delete item that is deleted on the server
                        //
                        status = new SyncOperationStatus[1];
                        cmd = (ModificationCommand)((SyncItemImpl)syncItemA).getPropertyValue(
                            SyncItemImpl.PROPERTY_COMMAND);
                        status[0] = new Sync4jOperationStatusOK(operation, sources[1], cmd);
                        break;
                    }
            }
        }

        return status;
    }

    /**
     * Check if there are some items to delete on the client.
     * If the client specifies a exclusive filter, the server has to delete
     * on the client all items that don't satisfy the filter. The server
     * knows the item on the client using the mapping. In order to do this,
     * we call the method getAllSyncItemKeys to retrieve all items in the filter
     * criteria. Then, we check all items in the mapping with this set; if an
     * item isn't in the set, we have to delete it on the client
     * @param allItemsKeys all items on the server inside the filter criteria
     * @param mapping the mapping
     * @param Bm the list of the changed items on the server
     */
    private void checkForItemsToDelete(int           syncFilterType,
                                       SyncItemKey[] allItemsKeys,
                                       ClientMapping mapping,
                                       List          Bm) {

        List guidluid = mapping.getMapping();

        if (syncFilterType == SYNC_WITH_EXCLUSIVE_FILTER) {

            if (log.isLoggable(Level.FINEST)) {
                log.finest("Check for items on the client (seeing the mapping) " +
                           "in order to discovery the items to delete");
            }

            List        keysList = Arrays.asList(allItemsKeys);
            SyncItemKey tmp      = new SyncItemKey("tmp");

            Iterator it = guidluid.iterator();
            String guid = null;
            ClientMappingEntry mappingEntry = null;

            while (it.hasNext()) {
                mappingEntry = ( (ClientMappingEntry)it.next());
                if (!mappingEntry.isValid()) {
                    continue;
                }
                guid = mappingEntry.getGuid();

                tmp.setKeyValue(guid);

                if (keysList.contains(tmp)) {
                    //
                    // The item satisfies the filter
                    //
                } else {

                    //
                    // We simulate the item on the server has been deleted
                    //
                    SyncItemImpl dummyItem = new SyncItemImpl(sources[0],
                        guid,
                        SyncItemState.DELETED);

                    //
                    // If there is already an item with the same key in Bm,
                    // the add isn't required (the item is already handled)
                    //
                    if (!Bm.contains(dummyItem)) {
                        if (log.isLoggable(Level.FINEST)) {
                            log.finest("The item '" + guid + "' is in the mapping but " +
                                       "it doesn't match the filter...we have to delete it");
                        }
                        Bm.add(dummyItem);
                    } else {
                        if (log.isLoggable(Level.FINEST)) {
                            log.finest("The item '" + guid + "' is in the mapping but " +
                                       " it doesn't match the filter...we have to delete" +
                                       " it but it's already in the Bm list");
                        }
                    }
                }
            }
        }
    }

    /**
     * Check if there are some items to add to the client because if the filter
     * changes some server have to added to the client.
     * So, an item could be not updated but the server has to send an Add command.
     * As borderline case, the client could not have specific any filter.
     * @param allItemsKeys all items on the server inside the filter criteria
     * @param mapping the mapping
     * @param Bm the list of the changed items on the server
     */
    private void checkForItemsToAdd(SyncItemKey[] allItemsKeys,
                                    ClientMapping mapping,
                                    List Bm) {

        List guidluid = mapping.getMapping();

        if (log.isLoggable(Level.FINEST)) {
            log.finest("Check the items that satisfy the filter criteria in" +
                       " order to discovery the items to add on the client");
        }

        Iterator itMapping = null;
        String   guid      = null;
        boolean  itemFound = false;
        ClientMappingEntry mappingEntry = null;

        for (int i = 0; i < allItemsKeys.length; i++) {

            itMapping = guidluid.iterator();

            while (itMapping.hasNext()) {
                itemFound = false;
                mappingEntry = (ClientMappingEntry)itMapping.next();
                if (!mappingEntry.isValid()) {
                    continue;
                }
                guid = mappingEntry.getGuid();
                if (guid.equals(allItemsKeys[i].getKeyAsString())) {
                    //
                    // The item is already in the mapping (so it's already
                    // on the client)
                    //
                    if (log.isLoggable(Level.FINEST)) {
                        log.finest("The item '" + guid + "' is already on the client");
                    }
                    itemFound = true;
                    break;
                }
            }
            if (!itemFound) {
                //
                // The item is on the server and it satisfies the filter
                // but it isn't in the mapping (so it isn't on the client).
                // The server has to send the item with an Add command.
                // In order to do this, we create a Dummy new item
                //

                SyncItemImpl dummyItem = new SyncItemImpl(sources[0],
                    allItemsKeys[i].getKeyValue(),
                    SyncItemState.NEW);

                //
                // If there is already an item with the same key in Bm,
                // the addition isn't required (the item is already handled)
                //
                if (!Bm.contains(dummyItem)) {
                    if (log.isLoggable(Level.FINEST)) {
                        log.finest("The item '" + allItemsKeys[i].getKeyAsString() +
                                   "' isn't on the client..." +
                                   " we have to add it");
                    }

                    Bm.add(dummyItem);
                } else {
                    if (log.isLoggable(Level.FINEST)) {
                        log.finest("The item '" + allItemsKeys[i].getKeyAsString() +
                                   "' isn't on the client" +
                                   " but it's already in the Bm list");
                    }
                }
            }
        }
    }

}

