/**
 * Copyright (C) 2005-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.funambol.server.notification;

import java.util.logging.Logger;

import com.funambol.framework.core.Alert;

import com.funambol.framework.logging.Sync4jLoggerName;
import com.funambol.framework.logging.Sync4jLogger;

import com.funambol.framework.notification.AbstractNotificationEngine;
import com.funambol.framework.notification.NotificationException;
import com.funambol.framework.notification.builder.DSNotificationMessageBuilder;
import com.funambol.framework.notification.sender.Sender;

import com.funambol.framework.server.Sync4jDevice;
import com.funambol.framework.server.inventory.DeviceInventoryException;

import com.funambol.framework.tools.beans.BeanException;

import com.funambol.server.config.Configuration;
import com.funambol.framework.filter.Clause;
import com.funambol.framework.filter.WhereClause;
import com.funambol.framework.security.Sync4jPrincipal;
import java.util.logging.Level;
import com.funambol.framework.security.*;
import com.funambol.framework.server.store.*;

/**
 * This class is an implementation of NotificationEngine interface
 *
 * @see NotificationEngine
 *
 * @version $Id: NotificationEngineImpl.java,v 1.2 2006/01/30 17:13:51 nichele Exp $
 */
public class NotificationEngineImpl extends AbstractNotificationEngine {

    // ------------------------------------------------------------ Private data
    private static final Logger log =
        Sync4jLogger.getLogger(Sync4jLoggerName.SERVER_DS_NOTIFICATION);

    // ------------------------------------------------------------- Constructor
    public NotificationEngineImpl(Configuration config) {
        super(config);
    }

    // ---------------------------------------------------------- Public Methods

    /**
     * Sends a notification message to the given device.
     *
     * @param deviceId device identifier
     * @param alerts   array of Alert command for the datastore to sync, used
     *                 to create the binary message to send
     * @param uimode   user intraction mode
     */
    public void sendNotificationMessage(String  deviceId,
                                        Alert[] alerts  ,
                                        int     uimode  )
    throws NotificationException {

        int sessionId = 0;
        DSNotificationMessageBuilder builder = null;
        Sender sender = null;

        if (log.isLoggable(Level.FINEST)) {
            log.finest("Try to notify device '" + deviceId + "'");
        }

        Sync4jDevice device = new Sync4jDevice(deviceId);
        boolean deviceFound = false;
        try {
            deviceFound = readDevice(device);
        } catch(DeviceInventoryException e) {
            throw new NotificationException("Error reading device '" + deviceId + "'");
        }
        if (!deviceFound) {
            throw new NotificationException("Device '" + deviceId + "' not found");
        }


        String notificationBuilder = device.getNotificationBuilder();
        String notificationSender  = device.getNotificationSender();


        if (notificationBuilder == null || notificationBuilder.equals("")) {
            if (log.isLoggable(Level.FINEST)) {
                log.finest("No notification builder configured for '" + deviceId +
                           "'");
            }
            return;
        }

        if (notificationSender == null || notificationSender.equals("")) {
            if (log.isLoggable(Level.FINEST)) {
                log.finest("No notification sender configured for '" + deviceId +
                           "'");
            }
            return;
        }


        try {
            builder = (DSNotificationMessageBuilder)
                getConfiguration().getBeanInstanceByName(notificationBuilder);
        } catch (BeanException e) {
            log.severe("Error creating the builder '" + notificationBuilder + "': " + e.getMessage());
            throw new NotificationException(e.getMessage());
        }

        try {
            sender = (Sender)
                getConfiguration().getBeanInstanceByName(notificationSender);
        } catch (BeanException e) {
            log.severe("Error creating the sender '" + notificationBuilder + "': " + e.getMessage());
            throw new NotificationException(e.getMessage());
        }

        try {
            String serverId =
                getConfiguration().getServerConfig().getServerInfo().getDevID();
            Float protocolVersion = new Float(1.2);

            //build message
            byte[] message = builder.buildMessage(sessionId                   ,
                                                  device.getAddress()         ,
                                                  serverId                    ,
                                                  device.getServerPassword()  ,
                                                  device.getServerNonce()     ,
                                                  alerts                      ,
                                                  uimode                      ,
                                                  protocolVersion.floatValue());
            //send message
            sender.sendNotificationMessage(device.getAddress(), message);
        } catch (NotificationException e) {
            log.throwing(this.getClass().getName(), "sendNotificationMessage", e);
            log.severe("Error sending message: " + e.getMessage());
            throw new NotificationException("Error sending message: " + e.getMessage());
        }
    }


    /**
     * Sends a notification message to all devices of the principals with the
     * given username.
     *
     * @param username the username
     * @param alerts   array of Alert command for the datastore to sync, used
     *                 to create the binary message to send
     * @param uimode   user intraction mode
     */
    public void sendNotificationMessages(String  username,
                                         Alert[] alerts  ,
                                         int     uimode  )
    throws NotificationException {

        Clause clause = new WhereClause("username",
                                        new String[] {username},
                                        WhereClause.OPT_EQ,
                                        true
                        );

        Sync4jPrincipal[] principals = null;
        try {
            principals = (Sync4jPrincipal[]) getStore().read(new
                    Sync4jPrincipal(), clause);
        } catch (PersistentStoreException e) {
            log.throwing(this.getClass().getName(), "sendNotificationMessages", e);
            log.severe("Error reading principals: " + e.getMessage());
            throw new NotificationException("Error reading principals: " + e.getMessage());
        }

         if (principals == null || principals.length == 0) {
             if (log.isLoggable(Level.FINEST)) {
                 log.finest("No principal found with username '" + username + "'");
                 log.finest("No devices to notify");
             }
         }

         int num = principals.length;

         for (int i=0; i<num; i++) {
            try {
                sendNotificationMessage(principals[i].getDeviceId(), alerts,
                                        uimode);
            } catch (Throwable ex) {
                //
                // Ignore it...we have to notify next device...
                //
            }
         }

    }
}
