/**
 * Copyright (C) 2003-2006 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.funambol.server.store;

import java.util.ArrayList;
import java.sql.*;

import com.funambol.framework.filter.*;

import com.funambol.framework.server.store.*;
import com.funambol.framework.security.Sync4jPrincipal;
import com.funambol.framework.tools.DBTools;

import com.funambol.server.tools.IdSpaceGenerator;

/**
 * This is the store for principal information of the Sync4j engine.
 * Currently it persistes the following classes:
 * <ul>
 * <li>com.funambol.framework.security.Sync4jPrincipal</li>
 * </ul>
 *
 * @version $Id: PrincipalPersistentStore.java,v 1.1 2006/01/18 17:00:12 luigiafassina Exp $
 */
public class PrincipalPersistentStore
extends BasePersistentStore
implements PersistentStore, java.io.Serializable {

    // --------------------------------------------------------------- Constants

    private static final String OPT_INSERT     = "INSERT";

    public static final int SQL_INSERT_PRINCIPAL             =  0;
    public static final int SQL_GET_PRINCIPAL                =  1;
    public static final int SQL_SELECT_PRINCIPAL             =  2;
    public static final int SQL_UPDATE_PRINCIPAL             =  3;
    public static final int SQL_SELECT_ALL_PRINCIPALS        =  4;
    public static final int SQL_DELETE_PRINCIPAL             =  5;
    public static final int SQL_DELETE_CLIENT_MAPPING        =  6;
    public static final int SQL_DELETE_LAST_SYNC             =  7;
    public static final int SQL_COUNT_PRINCIPALS             =  8;

    public static final String NS_PRINCIPAL = "principal";

    // -------------------------------------------------------------- Properties

    protected String[] sql = null;

    public void setSql(String[] sql) {
        this.sql = sql;
    }

    public String[] getSql() {
        return this.sql;
    }

    // ------------------------------------------------------------ Private data
    private static final String SEARCH_COUNT_PRINCIPALS = "SCP";

    // ------------------------------------------------------------ Constructors
    // ---------------------------------------------------------- Public methods

    /**
     * Delete the principal from the data store.
     *
     * @param o The object that must be a Sync4jPrincipal for this method
     *
     * @throws PersistentException in case of error reading the data store
     */
    public boolean delete(Object o) throws PersistentStoreException {
        if (o instanceof Sync4jPrincipal) {
            Sync4jPrincipal p = (Sync4jPrincipal) o;

            Connection conn = null;
            PreparedStatement stmt = null;

            try {
                conn = dataSource.getConnection();

                long pid = Long.parseLong(p.getId());

                stmt = conn.prepareStatement(sql[SQL_DELETE_CLIENT_MAPPING]);
                stmt.setLong(1, pid);
                stmt.executeUpdate();
                stmt.close();

                stmt = conn.prepareStatement(sql[SQL_DELETE_LAST_SYNC]);
                stmt.setLong(1, pid);
                stmt.executeUpdate();
                stmt.close();

                stmt = conn.prepareStatement(sql[SQL_DELETE_PRINCIPAL]);
                stmt.setLong(1, pid);
                stmt.executeUpdate();

            } catch (SQLException e) {
                throw new PersistentStoreException("Error deleting the principal " + p, e);
            } finally {
                DBTools.close(conn, stmt, null);
            }
            return true;
        }
        return false;
    }

    public boolean store(Object o, String operation) throws PersistentStoreException {
        return false;
    }

    public boolean store(Object o) throws PersistentStoreException {
        if (o instanceof Sync4jPrincipal) {
            Sync4jPrincipal p = (Sync4jPrincipal) o;

            Connection conn = null;
            PreparedStatement stmt = null;
            ResultSet rs = null;
            int n = 0;

            try {
                conn = dataSource.getConnection();

                if (p.getId() != null && !p.getId().equals("")) {
                    stmt = conn.prepareStatement(sql[SQL_UPDATE_PRINCIPAL]);
                    stmt.setString(1, p.getUsername());
                    stmt.setString(2, p.getDeviceId());
                    stmt.setLong(3, Long.parseLong(p.getId()));
                    n = stmt.executeUpdate();
                    stmt.close(); stmt = null;
                } else {
                    //check if the principal already exist: verify username-deviceid
                    stmt = conn.prepareStatement(sql[SQL_SELECT_PRINCIPAL]);
                    stmt.setString(1, p.getUsername());
                    stmt.setString(2, p.getDeviceId());
                    rs = stmt.executeQuery();

                    if (rs.next() == false) {
                        n = 0;
                    } else {
                        n = 1;
                        p.setId       (rs.getString(1));
                        p.setUsername (rs.getString(2));
                        p.setDeviceId (rs.getString(3));
                    }
                    rs.close(); rs = null;
                    stmt.close(); stmt = null;
                }

                if (n == 0) {
                    // The first time!!!
                    int principalId = getPrincipalId();
                    p.setId(String.valueOf(principalId));

                    stmt = conn.prepareStatement(sql[SQL_INSERT_PRINCIPAL]);
                    stmt.setLong(1, principalId);
                    stmt.setString(2, p.getUsername());
                    stmt.setString(3, p.getDeviceId());
                    stmt.executeUpdate();
                }
                return true;
            } catch (SQLException e) {
                throw new PersistentStoreException("Error storing principal " + p, e);
            } finally {
                DBTools.close(conn, stmt, rs);
            }
        }
        return false;
    }

    /**
     * Read the principal from the data store. If <i>getId()</i> returns null,
     * it tries to read the principal from username/device, otherwise throws id
     * is used for the lookup.
     * @param o The object that must be a Sync4jPrincipal for this method
     * @throws PersistentException in case of error reading the data store
     */
    public boolean read(Object o) throws PersistentStoreException {
        if (o instanceof Sync4jPrincipal) {
            Sync4jPrincipal p = (Sync4jPrincipal) o;

            Connection conn = null;
            PreparedStatement stmt = null;
            ResultSet rs = null;

            try {
                conn = dataSource.getConnection();

                if (p.getId() == null) {
                    stmt = conn.prepareStatement(sql[SQL_SELECT_PRINCIPAL]);
                    stmt.setString(1, p.getUsername());
                    stmt.setString(2, p.getDeviceId());
                } else {
                    stmt = conn.prepareStatement(sql[SQL_GET_PRINCIPAL]);
                    stmt.setLong(1, Long.parseLong(p.getId()));
                }
                rs = stmt.executeQuery();

                if (rs.next() == false) {
                    throw new NotFoundException("Principal not found for "
                    + p.toString());
                }
                p.setId       (String.valueOf(rs.getLong(1)));
                p.setUsername (rs.getString(2));
                p.setDeviceId (rs.getString(3));

            } catch (SQLException e) {
                throw new PersistentStoreException("Error reading principal " + p, e);
            } finally {
                DBTools.close(conn, stmt, rs);
            }
            return true;
        }
        return false;
    }

    public Object[] read(Class objClass) throws PersistentStoreException {
        return null;
    }

    /**
     * Read all for the Sync4jPrincipal.
     * @param o The object that must be a Sync4jPrincipal for this method
     * @param clause condition where for select
     *
     * @return Object[] array of Sync4jPrincipals
     */
    public Object[] read(Object o, Clause clause) throws PersistentStoreException {
        if (!(o instanceof Sync4jPrincipal)) {
            return null; // Not me!!
        }
        Connection conn = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;

        ArrayList ret = new ArrayList();

        try {
            conn = dataSource.getConnection();

            PreparedWhere where = sqlHelperClause.getPreparedWhere(clause);
            String query = sql[SQL_SELECT_ALL_PRINCIPALS];
            if (where.sql.length() > 0){
                query += " where " + where.sql;
            }
            stmt = conn.prepareStatement(query);
            for (int i=0; i<where.parameters.length; ++i) {
                stmt.setObject(i+1, where.parameters[i]);
            }
            rs = stmt.executeQuery();

            while (rs.next()) {
                ret.add(
                    new Sync4jPrincipal(
                        String.valueOf(rs.getLong(1)),
                        rs.getString(2),
                        rs.getString(3)));
            }

            return ret.toArray(new Sync4jPrincipal[ret.size()]);
        } catch (SQLException e) {
            throw new PersistentStoreException("Error reading principals", e);
        } finally {
            DBTools.close(conn, stmt, rs);
        }
    }



    /**
     * Count the number of principals that satisfy the conditions specified in input
     * @param o The object of the Sync4jPrincipal type.
     * @param clause The additional where clause.
     * @return int number of principals
     */
    public int count(Object o, Clause clause) throws PersistentStoreException {
        if (!(o instanceof Sync4jPrincipal)) {
            return -1; // Not me!
        }
        Connection conn = null;
        PreparedStatement stmt = null;
        ResultSet rs = null;
        int n = 0;

        try {
            conn = dataSource.getConnection();

            PreparedWhere where = sqlHelperClause.getPreparedWhere(clause);
            String query = sql[SQL_COUNT_PRINCIPALS];
            if (where.sql.length()>0) {
                query += " where " + where.sql;
            }
            stmt = conn.prepareStatement(query);
            for (int i=0; i<where.parameters.length; ++i) {
                stmt.setObject(i+1, where.parameters[i]);
            }
            rs = stmt.executeQuery();

            while (rs.next()) {
                n = rs.getInt(1);
            }
            return n;
        } catch (SQLException e) {
            throw new PersistentStoreException("Error reading count principals ", e);
        } finally {
            DBTools.close(conn, stmt, rs);
        }
    }

    //---------------------------------------------------------- Private Methods

    /**
     * Return the next principal id reading it by IdSpaceGenerator
     * @return int
     */
    private int getPrincipalId() {
        IdSpaceGenerator idGenerator = new IdSpaceGenerator(NS_PRINCIPAL);
        return Integer.parseInt(idGenerator.next());
    }
}

